import re
import sys
from functools import partial
from numbers import Number as numeric_type

import numpy as np
from more_itertools import first, mark_ends

from yt._typing import FieldKey
from yt.data_objects.construction_data_containers import YTCoveringGrid
from yt.data_objects.image_array import ImageArray
from yt.fields.derived_field import DerivedField
from yt.funcs import fix_axis, is_sequence, iter_fields, mylog, validate_moment
from yt.units import dimensions
from yt.units.unit_object import Unit  # type: ignore
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.math_utils import compute_stddev_image
from yt.utilities.on_demand_imports import _astropy
from yt.utilities.parallel_tools.parallel_analysis_interface import parallel_root_only
from yt.visualization.fixed_resolution import FixedResolutionBuffer, ParticleImageBuffer
from yt.visualization.particle_plots import (
    ParticleAxisAlignedDummyDataSource,
    ParticleDummyDataSource,
    ParticleOffAxisDummyDataSource,
)
from yt.visualization.volume_rendering.off_axis_projection import off_axis_projection


class UnitfulHDU:
    def __init__(self, hdu):
        self.hdu = hdu
        self.header = self.hdu.header
        self.name = self.header["BTYPE"]
        self.units = self.header["BUNIT"]
        self.shape = self.hdu.shape

    @property
    def data(self):
        return YTArray(self.hdu.data, self.units)

    def __repr__(self):
        im_shape = " x ".join(str(s) for s in self.shape)
        return f"FITSImage: {self.name} ({im_shape}, {self.units})"


class FITSImageData:
    def __init__(
        self,
        data,
        fields=None,
        length_unit=None,
        width=None,
        img_ctr=None,
        wcs=None,
        current_time=None,
        time_unit=None,
        mass_unit=None,
        velocity_unit=None,
        magnetic_unit=None,
        ds=None,
        unit_header=None,
        **kwargs,
    ):
        r"""Initialize a FITSImageData object.

        FITSImageData contains a collection of FITS ImageHDU instances and
        WCS information, along with units for each of the images. FITSImageData
        instances can be constructed from ImageArrays, NumPy arrays, dicts
        of such arrays, FixedResolutionBuffers, and YTCoveringGrids. The latter
        two are the most powerful because WCS information can be constructed
        automatically from their coordinates.

        Parameters
        ----------
        data : FixedResolutionBuffer or a YTCoveringGrid. Or, an
            ImageArray, an numpy.ndarray, or dict of such arrays
            The data to be made into a FITS image or images.
        fields : single string or list of strings, optional
            The field names for the data. If *fields* is none and *data* has
            keys, it will use these for the fields. If *data* is just a
            single array one field name must be specified.
        length_unit : string
            The units of the WCS coordinates and the length unit of the file.
            Defaults to the length unit of the dataset, if there is one, or
            "cm" if there is not.
        width : float or YTQuantity
            The width of the image. Either a single value or iterable of values.
            If a float, assumed to be in *units*. Only used if this information
            is not already provided by *data*.
        img_ctr : array_like or YTArray
            The center coordinates of the image. If a list or NumPy array,
            it is assumed to be in *units*. This will overwrite any center
            coordinates potentially provided by *data*. Default in other cases
            is [0.0]*(number of dimensions).
        wcs : `~astropy.wcs.WCS` instance, optional
            Supply an AstroPy WCS instance. Will override automatic WCS
            creation from FixedResolutionBuffers and YTCoveringGrids.
        current_time : float, tuple, or YTQuantity, optional
            The current time of the image(s). If not specified, one will
            be set from the dataset if there is one. If a float, it will
            be assumed to be in *time_unit* units.
        time_unit : string
            The default time units of the file. Defaults to "s".
        mass_unit : string
            The default time units of the file. Defaults to "g".
        velocity_unit : string
            The default velocity units of the file. Defaults to "cm/s".
        magnetic_unit : string
            The default magnetic units of the file. Defaults to "gauss".
        ds : `~yt.static_output.Dataset` instance, optional
            The dataset associated with the image(s), typically used
            to transfer metadata to the header(s). Does not need to be
            specified if *data* has a dataset as an attribute.

        Examples
        --------

        >>> # This example uses a FRB.
        >>> ds = load("sloshing_nomag2_hdf5_plt_cnt_0150")
        >>> prj = ds.proj(2, "kT", weight_field=("gas", "density"))
        >>> frb = prj.to_frb((0.5, "Mpc"), 800)
        >>> # This example just uses the FRB and puts the coords in kpc.
        >>> f_kpc = FITSImageData(
        ...     frb, fields="kT", length_unit="kpc", time_unit=(1.0, "Gyr")
        ... )
        >>> # This example specifies a specific WCS.
        >>> from astropy.wcs import WCS
        >>> w = WCS(naxis=self.dimensionality)
        >>> w.wcs.crval = [30.0, 45.0]  # RA, Dec in degrees
        >>> w.wcs.cunit = ["deg"] * 2
        >>> nx, ny = 800, 800
        >>> w.wcs.crpix = [0.5 * (nx + 1), 0.5 * (ny + 1)]
        >>> w.wcs.ctype = ["RA---TAN", "DEC--TAN"]
        >>> scale = 1.0 / 3600.0  # One arcsec per pixel
        >>> w.wcs.cdelt = [-scale, scale]
        >>> f_deg = FITSImageData(frb, fields="kT", wcs=w)
        >>> f_deg.writeto("temp.fits")
        """

        if fields is not None:
            fields = list(iter_fields(fields))

        if ds is None:
            ds = getattr(data, "ds", None)

        self.fields = []
        self.field_units = {}

        if unit_header is None:
            self._set_units(
                ds, [length_unit, mass_unit, time_unit, velocity_unit, magnetic_unit]
            )
        else:
            self._set_units_from_header(unit_header)

        wcs_unit = str(self.length_unit.units)

        self._fix_current_time(ds, current_time)

        if width is None:
            width = 1.0
        if isinstance(width, tuple):
            if ds is None:
                width = YTQuantity(width[0], width[1])
            else:
                width = ds.quan(width[0], width[1])

        exclude_fields = [
            "x",
            "y",
            "z",
            "px",
            "py",
            "pz",
            "pdx",
            "pdy",
            "pdz",
            "weight_field",
        ]

        if isinstance(data, _astropy.pyfits.PrimaryHDU):
            data = _astropy.pyfits.HDUList([data])

        if isinstance(data, _astropy.pyfits.HDUList):
            self.hdulist = data
            for hdu in data:
                self.fields.append(hdu.header["btype"])
                self.field_units[hdu.header["btype"]] = hdu.header["bunit"]

            self.shape = self.hdulist[0].shape
            self.dimensionality = len(self.shape)
            wcs_names = [key for key in self.hdulist[0].header if "WCSNAME" in key]
            for name in wcs_names:
                if name == "WCSNAME":
                    key = " "
                else:
                    key = name[-1]
                w = _astropy.pywcs.WCS(
                    header=self.hdulist[0].header, key=key, naxis=self.dimensionality
                )
                setattr(self, "wcs" + key.strip().lower(), w)

            return

        self.hdulist = _astropy.pyfits.HDUList()

        stddev = False
        if hasattr(data, "keys"):
            img_data = data
            if fields is None:
                fields = list(img_data.keys())
            if hasattr(data, "data_source"):
                stddev = getattr(data.data_source, "moment", 1) == 2
        elif isinstance(data, np.ndarray):
            if fields is None:
                mylog.warning(
                    "No field name given for this array. Calling it 'image_data'."
                )
                fn = "image_data"
                fields = [fn]
            else:
                fn = fields[0]
            img_data = {fn: data}

        for fd in fields:
            if isinstance(fd, tuple):
                fname = fd[1]
            elif isinstance(fd, DerivedField):
                fname = fd.name[1]
            else:
                fname = fd
            if stddev:
                fname += "_stddev"
            self.fields.append(fname)

        # Sanity checking names
        s = set()
        duplicates = {f for f in self.fields if f in s or s.add(f)}
        if len(duplicates) > 0:
            for i, fd in enumerate(self.fields):
                if fd in duplicates:
                    if isinstance(fields[i], tuple):
                        ftype, fname = fields[i]
                    elif isinstance(fields[i], DerivedField):
                        ftype, fname = fields[i].name
                    else:
                        raise RuntimeError(
                            f"Cannot distinguish between fields with same name {fd}!"
                        )
                    self.fields[i] = f"{ftype}_{fname}"

        for is_first, _is_last, (i, (name, field)) in mark_ends(
            enumerate(zip(self.fields, fields))
        ):
            if name not in exclude_fields:
                this_img = img_data[field]
                if hasattr(img_data[field], "units"):
                    has_code_unit = False
                    for atom in this_img.units.expr.atoms():
                        if str(atom).startswith("code"):
                            has_code_unit = True
                    if has_code_unit:
                        mylog.warning(
                            "Cannot generate an image with code "
                            "units. Converting to units in CGS."
                        )
                        funits = this_img.units.get_base_equivalent("cgs")
                        this_img.convert_to_units(funits)
                    else:
                        funits = this_img.units
                    self.field_units[name] = str(funits)
                else:
                    self.field_units[name] = "dimensionless"
                mylog.info("Making a FITS image of field %s", name)
                if isinstance(this_img, ImageArray):
                    if i == 0:
                        self.shape = this_img.shape[::-1]
                    this_img = np.asarray(this_img)
                else:
                    if i == 0:
                        self.shape = this_img.shape
                    this_img = np.asarray(this_img.T)
                if is_first:
                    hdu = _astropy.pyfits.PrimaryHDU(this_img)
                else:
                    hdu = _astropy.pyfits.ImageHDU(this_img)
                hdu.name = name
                hdu.header["btype"] = name
                hdu.header["bunit"] = re.sub("()", "", self.field_units[name])
                for unit in ("length", "time", "mass", "velocity", "magnetic"):
                    if unit == "magnetic":
                        short_unit = "bf"
                    else:
                        short_unit = unit[0]
                    key = f"{short_unit}unit"
                    value = getattr(self, f"{unit}_unit")
                    if value is not None:
                        hdu.header[key] = float(value.value)
                        hdu.header.comments[key] = f"[{value.units}]"
                hdu.header["time"] = float(self.current_time.value)
                if hasattr(self, "current_redshift"):
                    hdu.header["HUBBLE"] = self.hubble_constant
                    hdu.header["REDSHIFT"] = self.current_redshift
                self.hdulist.append(hdu)

        self.dimensionality = len(self.shape)

        if wcs is None:
            w = _astropy.pywcs.WCS(
                header=self.hdulist[0].header, naxis=self.dimensionality
            )
            # FRBs and covering grids are special cases where
            # we have coordinate information, so we take advantage
            # of this and construct the WCS object
            if isinstance(img_data, FixedResolutionBuffer):
                dx = (img_data.bounds[1] - img_data.bounds[0]).to_value(wcs_unit)
                dy = (img_data.bounds[3] - img_data.bounds[2]).to_value(wcs_unit)
                dx /= self.shape[0]
                dy /= self.shape[1]
                if img_ctr is not None:
                    xctr, yctr = img_ctr
                else:
                    xctr = 0.5 * (img_data.bounds[1] + img_data.bounds[0]).to_value(
                        wcs_unit
                    )
                    yctr = 0.5 * (img_data.bounds[3] + img_data.bounds[2]).to_value(
                        wcs_unit
                    )
                center = [xctr, yctr]
                cdelt = [dx, dy]
            elif isinstance(img_data, YTCoveringGrid):
                cdelt = img_data.dds.to_value(wcs_unit)
                if img_ctr is not None:
                    center = img_ctr
                else:
                    center = 0.5 * (img_data.left_edge + img_data.right_edge).to_value(
                        wcs_unit
                    )
            else:
                # If img_data is just an array we use the width and img_ctr
                # parameters to determine the cell widths
                if img_ctr is None:
                    img_ctr = np.zeros(3)
                if not is_sequence(width):
                    width = [width] * self.dimensionality
                if isinstance(width[0], YTQuantity):
                    cdelt = [
                        wh.to_value(wcs_unit) / n for wh, n in zip(width, self.shape)
                    ]
                else:
                    cdelt = [float(wh) / n for wh, n in zip(width, self.shape)]
                center = img_ctr[: self.dimensionality]
            w.wcs.crpix = 0.5 * (np.array(self.shape) + 1)
            w.wcs.crval = center
            w.wcs.cdelt = cdelt
            w.wcs.ctype = ["linear"] * self.dimensionality
            w.wcs.cunit = [wcs_unit] * self.dimensionality
            self.set_wcs(w)
        else:
            self.set_wcs(wcs)

    def _fix_current_time(self, ds, current_time):
        if ds is None:
            registry = None
        else:
            registry = ds.unit_registry
        tunit = Unit(self.time_unit, registry=registry)
        if current_time is None:
            if ds is not None:
                current_time = ds.current_time
            else:
                self.current_time = YTQuantity(0.0, "s")
                return
        elif isinstance(current_time, numeric_type):
            current_time = YTQuantity(current_time, tunit)
        elif isinstance(current_time, tuple):
            current_time = YTQuantity(current_time[0], current_time[1])
        self.current_time = current_time.to(tunit)

    def _set_units(self, ds, base_units):
        if ds is not None:
            if getattr(ds, "cosmological_simulation", False):
                self.hubble_constant = ds.hubble_constant
                self.current_redshift = ds.current_redshift
        attrs = (
            "length_unit",
            "mass_unit",
            "time_unit",
            "velocity_unit",
            "magnetic_unit",
        )
        cgs_units = ("cm", "g", "s", "cm/s", "gauss")
        for unit, attr, cgs_unit in zip(base_units, attrs, cgs_units):
            if unit is None:
                if ds is not None:
                    u = getattr(ds, attr, None)
                elif attr == "velocity_unit":
                    u = self.length_unit / self.time_unit
                elif attr == "magnetic_unit":
                    u = np.sqrt(
                        4.0
                        * np.pi
                        * self.mass_unit
                        / (self.time_unit**2 * self.length_unit)
                    )
                else:
                    u = cgs_unit
            else:
                u = unit

            if isinstance(u, str):
                uq = YTQuantity(1.0, u)
            elif isinstance(u, numeric_type):
                uq = YTQuantity(u, cgs_unit)
            elif isinstance(u, YTQuantity):
                uq = u.copy()
            elif isinstance(u, tuple):
                uq = YTQuantity(u[0], u[1])
            else:
                uq = None

            if uq is not None:
                atoms = {str(a) for a in uq.units.expr.atoms()}
                if hasattr(self, "hubble_constant"):
                    # Don't store cosmology units
                    if str(uq.units).startswith("cm") or "h" in atoms or "a" in atoms:
                        uq.convert_to_cgs()
                if any(a.startswith("code") for a in atoms):
                    # Don't store code units
                    mylog.warning(
                        "Cannot use code units of '%s' "
                        "when creating a FITSImageData instance! "
                        "Converting to a cgs equivalent.",
                        uq.units,
                    )
                    uq.convert_to_cgs()

            if attr == "length_unit" and uq.value != 1.0:
                mylog.warning("Converting length units from %s to %s.", uq, uq.units)
                uq = YTQuantity(1.0, uq.units)

            setattr(self, attr, uq)

    def _set_units_from_header(self, header):
        if "hubble" in header:
            self.hubble_constant = header["HUBBLE"]
            self.current_redshift = header["REDSHIFT"]
        for unit in ["length", "time", "mass", "velocity", "magnetic"]:
            if unit == "magnetic":
                key = "BFUNIT"
            else:
                key = unit[0].upper() + "UNIT"
            if key not in header:
                continue
            u = header.comments[key].strip("[]")
            uq = YTQuantity(header[key], u)
            setattr(self, unit + "_unit", uq)

    def set_wcs(self, wcs, wcsname=None, suffix=None):
        """
        Set the WCS coordinate information for all images
        with a WCS object *wcs*.
        """
        if wcsname is None:
            wcs.wcs.name = "yt"
        else:
            wcs.wcs.name = wcsname
        h = wcs.to_header()
        if suffix is None:
            self.wcs = wcs
        else:
            setattr(self, "wcs" + suffix, wcs)
        for img in self.hdulist:
            for k, v in h.items():
                kk = k
                if suffix is not None:
                    kk += suffix
                img.header[kk] = v

    def change_image_name(self, old_name, new_name):
        """
        Change the name of a FITS image.

        Parameters
        ----------
        old_name : string
            The old name of the image.
        new_name : string
            The new name of the image.
        """
        idx = self.fields.index(old_name)
        self.hdulist[idx].name = new_name
        self.hdulist[idx].header["BTYPE"] = new_name
        self.field_units[new_name] = self.field_units.pop(old_name)
        self.fields[idx] = new_name

    def convolve(self, field, kernel, **kwargs):
        """
        Convolve an image with a kernel, either a simple
        Gaussian kernel or one provided by AstroPy. Currently,
        this only works for 2D images.

        All keyword arguments are passed to
        :meth:`~astropy.convolution.convolve`.

        Parameters
        ----------
        field : string
            The name of the field to convolve.
        kernel : float, YTQuantity, (value, unit) tuple, or AstroPy Kernel object
            The kernel to convolve the image with. If this is an AstroPy Kernel
            object, the image will be convolved with it. Otherwise, it is
            assumed that the kernel is a Gaussian and that this value is
            the standard deviation. If a float, it is assumed that the units
            are pixels, but a (value, unit) tuple or YTQuantity can be supplied
            to specify the standard deviation in physical units.

        Examples
        --------
        >>> fid = FITSSlice(ds, "z", ("gas", "density"))
        >>> fid.convolve("density", (3.0, "kpc"))
        """
        if self.dimensionality == 3:
            raise RuntimeError("Convolution currently only works for 2D FITSImageData!")
        conv = _astropy.conv
        if field not in self.keys():
            raise KeyError(f"{field} not an image!")
        idx = self.fields.index(field)
        if not isinstance(kernel, conv.Kernel):
            if not isinstance(kernel, numeric_type):
                unit = str(self.wcs.wcs.cunit[0])
                pix_scale = YTQuantity(self.wcs.wcs.cdelt[0], unit)
                if isinstance(kernel, tuple):
                    stddev = YTQuantity(kernel[0], kernel[1]).to(unit)
                else:
                    stddev = kernel.to(unit)
                kernel = stddev / pix_scale
            kernel = conv.Gaussian2DKernel(x_stddev=kernel)
        self.hdulist[idx].data = conv.convolve(self.hdulist[idx].data, kernel, **kwargs)

    def update_header(self, field, key, value):
        """
        Update the FITS header for *field* with a
        *key*, *value* pair. If *field* == "all", all
        headers will be updated.
        """
        if field == "all":
            for img in self.hdulist:
                img.header[key] = value
        else:
            if field not in self.keys():
                raise KeyError(f"{field} not an image!")
            idx = self.fields.index(field)
            self.hdulist[idx].header[key] = value

    def keys(self):
        return self.fields

    def has_key(self, key):
        return key in self.fields

    def values(self):
        return [self[k] for k in self.fields]

    def items(self):
        return [(k, self[k]) for k in self.fields]

    def __getitem__(self, item):
        return UnitfulHDU(self.hdulist[item])

    def __repr__(self):
        return str([self[k] for k in self.keys()])

    def info(self, output=None):
        """
        Summarize the info of the HDUs in this `FITSImageData`
        instance.

        Note that this function prints its results to the console---it
        does not return a value.

        Parameters
        ----------
        output : file, boolean, optional
            A file-like object to write the output to.  If `False`, does not
            output to a file and instead returns a list of tuples representing
            the FITSImageData info.  Writes to ``sys.stdout`` by default.
        """
        hinfo = self.hdulist.info(output=False)
        num_cols = len(hinfo[0])
        if output is None:
            output = sys.stdout
        if num_cols == 8:
            header = "No.    Name      Ver    Type      Cards   Dimensions   Format     Units"
            format = "{:3d}  {:10}  {:3} {:11}  {:5d}   {}   {}   {}"
        else:
            header = (
                "No.    Name         Type      Cards   Dimensions   Format     Units"
            )
            format = "{:3d}  {:10}  {:11}  {:5d}   {}   {}   {}"
        if self.hdulist._file is None:
            name = "(No file associated with this FITSImageData)"
        else:
            name = self.hdulist._file.name
        results = [f"Filename: {name}", header]
        for line in hinfo:
            units = self.field_units[self.hdulist[line[0]].header["btype"]]
            summary = tuple(list(line[:-1]) + [units])
            if output:
                results.append(format.format(*summary))
            else:
                results.append(summary)

        if output:
            output.write("\n".join(results))
            output.write("\n")
            output.flush()
        else:
            return results[2:]

    @parallel_root_only
    def writeto(self, fileobj, fields=None, overwrite=False, **kwargs):
        r"""
        Write all of the fields or a subset of them to a FITS file.

        Parameters
        ----------
        fileobj : string
            The name of the file to write to.
        fields : list of strings, optional
            The fields to write to the file. If not specified
            all of the fields in the buffer will be written.
        overwrite : boolean
            Whether or not to overwrite a previously existing file.
            Default: False
        **kwargs
            Additional keyword arguments are passed to
            :meth:`~astropy.io.fits.HDUList.writeto`.
        """
        if fields is None:
            hdus = self.hdulist
        else:
            hdus = _astropy.pyfits.HDUList()
            for field in fields:
                hdus.append(self.hdulist[field])
        hdus.writeto(fileobj, overwrite=overwrite, **kwargs)

    def to_glue(self, label="yt", data_collection=None):
        """
        Takes the data in the FITSImageData instance and exports it to
        Glue (http://glueviz.org) for interactive analysis. Optionally
        add a *label*. If you are already within the Glue environment, you
        can pass a *data_collection* object, otherwise Glue will be started.
        """
        from glue.core import Data, DataCollection
        from glue.core.coordinates import coordinates_from_header

        try:
            from glue.app.qt.application import GlueApplication
        except ImportError:
            from glue.qt.glue_application import GlueApplication

        image = Data(label=label)
        image.coords = coordinates_from_header(self.wcs.to_header())
        for k in self.fields:
            image.add_component(self[k].data, k)
        if data_collection is None:
            dc = DataCollection([image])
            app = GlueApplication(dc)
            app.start()
        else:
            data_collection.append(image)

    def to_aplpy(self, **kwargs):
        """
        Use APLpy (http://aplpy.github.io) for plotting. Returns an
        `aplpy.FITSFigure` instance. All keyword arguments are passed
        to the `aplpy.FITSFigure` constructor.
        """
        import aplpy

        return aplpy.FITSFigure(self.hdulist, **kwargs)

    def get_data(self, field):
        """
        Return the data array of the image corresponding to *field*
        with units attached. Deprecated.
        """
        return self[field].data

    def set_unit(self, field, units):
        """
        Set the units of *field* to *units*.
        """
        if field not in self.keys():
            raise KeyError(f"{field} not an image!")
        idx = self.fields.index(field)
        new_data = YTArray(self.hdulist[idx].data, self.field_units[field]).to(units)
        self.hdulist[idx].data = new_data.v
        self.hdulist[idx].header["bunit"] = units
        self.field_units[field] = units

    def pop(self, key):
        """
        Remove a field with name *key*
        and return it as a new FITSImageData
        instance.
        """
        if key not in self.keys():
            raise KeyError(f"{key} not an image!")
        idx = self.fields.index(key)
        im = self.hdulist.pop(idx)
        self.field_units.pop(key)
        self.fields.remove(key)
        f = _astropy.pyfits.PrimaryHDU(im.data, header=im.header)
        return FITSImageData(f, current_time=f.header["TIME"], unit_header=f.header)

    def close(self):
        self.hdulist.close()

    @classmethod
    def from_file(cls, filename):
        """
        Generate a FITSImageData instance from one previously written to
        disk.

        Parameters
        ----------
        filename : string
            The name of the file to open.
        """
        f = _astropy.pyfits.open(filename, lazy_load_hdus=False)
        return cls(f, current_time=f[0].header["TIME"], unit_header=f[0].header)

    @classmethod
    def from_images(cls, image_list):
        """
        Generate a new FITSImageData instance from a list of FITSImageData
        instances.

        Parameters
        ----------
        image_list : list of FITSImageData instances
            The images to be combined.
        """
        image_list = image_list if isinstance(image_list, list) else [image_list]
        first_image = first(image_list)

        w = first_image.wcs
        img_shape = first_image.shape
        data = []
        for fid in image_list:
            assert_same_wcs(w, fid.wcs)
            if img_shape != fid.shape:
                raise RuntimeError("Images do not have the same shape!")
            for hdu in fid.hdulist:
                if len(data) == 0:
                    data.append(_astropy.pyfits.PrimaryHDU(hdu.data, header=hdu.header))
                else:
                    data.append(_astropy.pyfits.ImageHDU(hdu.data, header=hdu.header))
        data = _astropy.pyfits.HDUList(data)
        return cls(
            data,
            current_time=first_image.current_time,
            unit_header=first_image[0].header,
        )

    def create_sky_wcs(
        self,
        sky_center,
        sky_scale,
        ctype=None,
        crota=None,
        cd=None,
        pc=None,
        wcsname="celestial",
        replace_old_wcs=True,
    ):
        """
        Takes a Cartesian WCS and converts it to one in a
        sky-based coordinate system.

        Parameters
        ----------
        sky_center : iterable of floats
            Reference coordinates of the WCS in degrees.
        sky_scale : tuple or YTQuantity
            Conversion between an angle unit and a length unit,
            e.g. (3.0, "arcsec/kpc")
        ctype : list of strings, optional
            The type of the coordinate system to create. Default:
            A "tangential" projection.
        crota : 2-element ndarray, optional
            Rotation angles between cartesian coordinates and
            the celestial coordinates.
        cd : 2x2-element ndarray, optional
            Dimensioned coordinate transformation matrix.
        pc : 2x2-element ndarray, optional
            Coordinate transformation matrix.
        wcsname : string, optional
            The name of the WCS to be stored in the FITS header.
        replace_old_wcs : boolean, optional
            If True, the original WCS will be overwritten but
            first copied to a second WCS ("WCSAXESA"). If False, this
            new WCS will be placed into the second WCS.
        """
        if ctype is None:
            ctype = ["RA---TAN", "DEC--TAN"]
        old_wcs = self.wcs
        naxis = old_wcs.naxis
        crval = [sky_center[0], sky_center[1]]
        if isinstance(sky_scale, YTQuantity):
            scaleq = sky_scale
        else:
            scaleq = YTQuantity(sky_scale[0], sky_scale[1])
        if scaleq.units.dimensions != dimensions.angle / dimensions.length:
            raise RuntimeError(
                f"sky_scale {sky_scale} not in correct dimensions of angle/length!"
            )
        deltas = old_wcs.wcs.cdelt
        units = [str(unit) for unit in old_wcs.wcs.cunit]
        new_dx = (YTQuantity(-deltas[0], units[0]) * scaleq).in_units("deg")
        new_dy = (YTQuantity(deltas[1], units[1]) * scaleq).in_units("deg")
        new_wcs = _astropy.pywcs.WCS(naxis=naxis)
        cdelt = [new_dx.v, new_dy.v]
        cunit = ["deg"] * 2
        if naxis == 3:
            crval.append(old_wcs.wcs.crval[2])
            cdelt.append(old_wcs.wcs.cdelt[2])
            ctype.append(old_wcs.wcs.ctype[2])
            cunit.append(old_wcs.wcs.cunit[2])
        new_wcs.wcs.crpix = old_wcs.wcs.crpix
        new_wcs.wcs.cdelt = cdelt
        new_wcs.wcs.crval = crval
        new_wcs.wcs.cunit = cunit
        new_wcs.wcs.ctype = ctype
        if crota is not None:
            new_wcs.wcs.crota = crota
        if cd is not None:
            new_wcs.wcs.cd = cd
        if pc is not None:
            new_wcs.wcs.cd = pc
        if replace_old_wcs:
            self.set_wcs(new_wcs, wcsname=wcsname)
            self.set_wcs(old_wcs, wcsname="yt", suffix="a")
        else:
            self.set_wcs(new_wcs, wcsname=wcsname, suffix="a")


class FITSImageBuffer(FITSImageData):
    pass


def sanitize_fits_unit(unit):
    if unit == "Mpc":
        mylog.info("Changing FITS file length unit to kpc.")
        unit = "kpc"
    elif unit == "au":
        unit = "AU"
    return unit


# This list allows one to determine which axes are the
# correct axes of the image in a right-handed coordinate
# system depending on which axis is sliced or projected
axis_wcs = [[1, 2], [2, 0], [0, 1]]


def construct_image(
    ds, axis, data_source, center, image_res, width, length_unit, origin="domain"
):
    if width is None:
        width = ds.domain_width[axis_wcs[axis]]
        unit = ds.get_smallest_appropriate_unit(width[0])
        mylog.info(
            "Making an image of the entire domain, "
            "so setting the center to the domain center."
        )
    else:
        width = ds.coordinates.sanitize_width(axis, width, None)
        unit = str(width[0].units)
    if is_sequence(image_res):
        nx, ny = image_res
    else:
        nx, ny = image_res, image_res
    dx = width[0] / nx
    dy = width[1] / ny
    crpix = [0.5 * (nx + 1), 0.5 * (ny + 1)]
    if unit == "unitary":
        unit = ds.get_smallest_appropriate_unit(ds.domain_width.max())
    elif unit == "code_length":
        unit = ds.get_smallest_appropriate_unit(ds.quan(1.0, "code_length"))
    unit = sanitize_fits_unit(unit)
    if length_unit is None:
        length_unit = unit
    if any(char.isdigit() for char in length_unit) and "*" in length_unit:
        length_unit = length_unit.split("*")[-1]
    cunit = [length_unit] * 2
    ctype = ["LINEAR"] * 2
    cdelt = [dx.in_units(length_unit), dy.in_units(length_unit)]
    if origin == "domain":
        if is_sequence(axis):
            crval = center.in_units(length_unit)
        else:
            crval = [center[idx].in_units(length_unit) for idx in axis_wcs[axis]]
    elif origin == "image":
        crval = np.zeros(2)
    if hasattr(data_source, "to_frb"):
        if is_sequence(axis):
            frb = data_source.to_frb(width[0], (nx, ny), height=width[1])
        else:
            frb = data_source.to_frb(width[0], (nx, ny), center=center, height=width[1])
    elif isinstance(data_source, ParticleDummyDataSource):
        if hasattr(data_source, "normal_vector"):
            # If we have a normal vector, this means
            # that the data source is off-axis
            bounds = (-width[0] / 2, width[0] / 2, -width[1] / 2, width[1] / 2)
            periodic = False
        else:
            # Otherwise, this is an on-axis data source
            axes = axis_wcs[axis]
            bounds = (
                center[axes[0]] - width[0] / 2,
                center[axes[0]] + width[0] / 2,
                center[axes[1]] - width[1] / 2,
                center[axes[1]] + width[1] / 2,
            )
            periodic = all(ds.periodicity)
        frb = ParticleImageBuffer(data_source, bounds, (nx, ny), periodic=periodic)
    else:
        frb = None
    w = _astropy.pywcs.WCS(naxis=2)
    w.wcs.crpix = crpix
    w.wcs.cdelt = cdelt
    w.wcs.crval = crval
    w.wcs.cunit = cunit
    w.wcs.ctype = ctype
    return w, frb, length_unit


def assert_same_wcs(wcs1, wcs2):
    from numpy.testing import assert_allclose

    assert wcs1.naxis == wcs2.naxis
    for i in range(wcs1.naxis):
        assert wcs1.wcs.cunit[i] == wcs2.wcs.cunit[i]
        assert wcs1.wcs.ctype[i] == wcs2.wcs.ctype[i]
    assert_allclose(wcs1.wcs.crpix, wcs2.wcs.crpix)
    assert_allclose(wcs1.wcs.cdelt, wcs2.wcs.cdelt)
    assert_allclose(wcs1.wcs.crval, wcs2.wcs.crval)
    crota1 = getattr(wcs1.wcs, "crota", None)
    crota2 = getattr(wcs2.wcs, "crota", None)
    if crota1 is None or crota2 is None:
        assert crota1 == crota2
    else:
        assert_allclose(wcs1.wcs.crota, wcs2.wcs.crota)
    cd1 = getattr(wcs1.wcs, "cd", None)
    cd2 = getattr(wcs2.wcs, "cd", None)
    if cd1 is None or cd2 is None:
        assert cd1 == cd2
    else:
        assert_allclose(wcs1.wcs.cd, wcs2.wcs.cd)
    pc1 = getattr(wcs1.wcs, "pc", None)
    pc2 = getattr(wcs2.wcs, "pc", None)
    if pc1 is None or pc2 is None:
        assert pc1 == pc2
    else:
        assert_allclose(wcs1.wcs.pc, wcs2.wcs.pc)


class FITSSlice(FITSImageData):
    r"""
    Generate a FITSImageData of an on-axis slice.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        The dataset object.
    axis : character or integer
        The axis of the slice. One of "x","y","z", or 0,1,2.
    fields : string or list of strings
        The fields to slice
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : 'center', 'c', 'left', 'l', 'right', 'r', id of a global extremum, or array-like
        The coordinate of the selection's center.
        Defaults to the 'center', i.e. center of the domain.

        Centering on the min or max of a field is supported by passing a tuple
        such as ('min', ('gas', 'density')) or ('max', ('gas', 'temperature'). A
        single string may also be used (e.g. "min_density" or
        "max_temperature"), though it's not as flexible and does not allow to
        select an exact field/particle type. With this syntax, the first field
        matching the provided name is selected.
        'max' or 'm' can be used as a shortcut for ('max', ('gas', 'density'))
        'min' can be used as a shortcut for ('min', ('gas', 'density'))

        One can also select an exact point as a 3 element coordinate sequence,
        e.g. [0.5, 0.5, 0]
        Units can be specified by passing in *center* as a tuple containing a
        3-element coordinate sequence and string unit name, e.g. ([0, 0.5, 0.5], "cm"),
        or by passing in a YTArray. Code units are assumed if unspecified.

        The domain edges along the selected *axis* can be selected with
        'left'/'l' and 'right'/'r' respectively.

    width : tuple or a float.
        Width can have four different formats to support variable x and y
        widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs wide
        in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a width
        that is 10 kiloparsecs wide along the x axis and 15 kiloparsecs wide
        along the y axis.  In the other two examples, code units are assumed,
        for example (0.2, 0.3) specifies a width that has an x width of 0.2 and
        a y width of 0.3 in code units.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    origin : string
        The origin of the coordinate system in the file. If "domain", then the
        center coordinates will be the same as the center of the image as
        defined by the *center* keyword argument. If "image", then the center
        coordinates will be set to (0,0). Default: "domain"
    """

    def __init__(
        self,
        ds,
        axis,
        fields,
        image_res=512,
        center="center",
        width=None,
        length_unit=None,
        *,
        origin="domain",
        **kwargs,
    ):
        fields = list(iter_fields(fields))
        axis = fix_axis(axis, ds)
        center, dcenter = ds.coordinates.sanitize_center(center, axis)
        slc = ds.slice(axis, center[axis], **kwargs)
        w, frb, lunit = construct_image(
            ds,
            axis,
            slc,
            dcenter,
            image_res,
            width,
            length_unit,
            origin=origin,
        )
        super().__init__(frb, fields=fields, length_unit=lunit, wcs=w)


class FITSProjection(FITSImageData):
    r"""
    Generate a FITSImageData of an on-axis projection.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        The dataset object.
    axis : character or integer
        The axis along which to project. One of "x","y","z", or 0,1,2.
    fields : string or list of strings
        The fields to project
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : 'center', 'c', 'left', 'l', 'right', 'r', id of a global extremum, or array-like
        The coordinate of the selection's center.
        Defaults to the 'center', i.e. center of the domain.

        Centering on the min or max of a field is supported by passing a tuple
        such as ('min', ('gas', 'density')) or ('max', ('gas', 'temperature'). A
        single string may also be used (e.g. "min_density" or
        "max_temperature"), though it's not as flexible and does not allow to
        select an exact field/particle type. With this syntax, the first field
        matching the provided name is selected.
        'max' or 'm' can be used as a shortcut for ('max', ('gas', 'density'))
        'min' can be used as a shortcut for ('min', ('gas', 'density'))

        One can also select an exact point as a 3 element coordinate sequence,
        e.g. [0.5, 0.5, 0]
        Units can be specified by passing in *center* as a tuple containing a
        3-element coordinate sequence and string unit name, e.g. ([0, 0.5, 0.5], "cm"),
        or by passing in a YTArray. Code units are assumed if unspecified.

        The domain edges along the selected *axis* can be selected with
        'left'/'l' and 'right'/'r' respectively.
    width : tuple or a float.
        Width can have four different formats to support variable
        x and y widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs
        wide in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a
        width that is 10 kiloparsecs wide along the x-axis and 15
        kiloparsecs wide along the y-axis.  In the other two examples, code
        units are assumed, for example (0.2, 0.3) specifies a width that has an
        x width of 0.2 and a y width of 0.3 in code units.
    weight_field : string
        The field used to weight the projection.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    origin : string
        The origin of the coordinate system in the file. If "domain", then the
        center coordinates will be the same as the center of the image as
        defined by the *center* keyword argument. If "image", then the center
        coordinates will be set to (0,0). Default: "domain"
    moment : integer, optional
        for a weighted projection, moment = 1 (the default) corresponds to a
        weighted average. moment = 2 corresponds to a weighted standard
        deviation.
    """

    def __init__(
        self,
        ds,
        axis,
        fields,
        image_res=512,
        center="center",
        width=None,
        weight_field=None,
        length_unit=None,
        *,
        origin="domain",
        moment=1,
        **kwargs,
    ):
        fields = list(iter_fields(fields))
        axis = fix_axis(axis, ds)
        center, dcenter = ds.coordinates.sanitize_center(center, axis)
        prj = ds.proj(
            fields[0], axis, weight_field=weight_field, moment=moment, **kwargs
        )
        w, frb, lunit = construct_image(
            ds,
            axis,
            prj,
            dcenter,
            image_res,
            width,
            length_unit,
            origin=origin,
        )
        super().__init__(frb, fields=fields, length_unit=lunit, wcs=w)


class FITSParticleProjection(FITSImageData):
    r"""
    Generate a FITSImageData of an on-axis projection of a
    particle field.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        The dataset object.
    axis : character or integer
        The axis along which to project. One of "x","y","z", or 0,1,2.
    fields : string or list of strings
        The fields to project
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : 'center', 'c', 'left', 'l', 'right', 'r', id of a global extremum, or array-like
        The coordinate of the selection's center.
        Defaults to the 'center', i.e. center of the domain.

        Centering on the min or max of a field is supported by passing a tuple
        such as ('min', ('gas', 'density')) or ('max', ('gas', 'temperature'). A
        single string may also be used (e.g. "min_density" or
        "max_temperature"), though it's not as flexible and does not allow to
        select an exact field/particle type. With this syntax, the first field
        matching the provided name is selected.
        'max' or 'm' can be used as a shortcut for ('max', ('gas', 'density'))
        'min' can be used as a shortcut for ('min', ('gas', 'density'))

        One can also select an exact point as a 3 element coordinate sequence,
        e.g. [0.5, 0.5, 0]
        Units can be specified by passing in *center* as a tuple containing a
        3-element coordinate sequence and string unit name, e.g. ([0, 0.5, 0.5], "cm"),
        or by passing in a YTArray. Code units are assumed if unspecified.

        The domain edges along the selected *axis* can be selected with
        'left'/'l' and 'right'/'r' respectively.
    width : tuple or a float.
        Width can have four different formats to support variable
        x and y widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs
        wide in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a
        width that is 10 kiloparsecs wide along the x-axis and 15
        kiloparsecs wide along the y-axis.  In the other two examples, code
        units are assumed, for example (0.2, 0.3) specifies a width that has an
        x width of 0.2 and a y width of 0.3 in code units.
    depth : A tuple or a float
         A tuple containing the depth to project through and the string
         key of the unit: (width, 'unit').  If set to a float, code units
         are assumed. Defaults to the entire domain.
    weight_field : string
        The field used to weight the projection.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    deposition : string, optional
        Controls the order of the interpolation of the particles onto the
        mesh. "ngp" is 0th-order "nearest-grid-point" method (the default),
        "cic" is 1st-order "cloud-in-cell".
    density : boolean, optional
        If True, the quantity to be projected will be divided by the area of
        the cells, to make a projected density of the quantity. Default: False
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source : yt.data_objects.data_containers.YTSelectionContainer, optional
        If specified, this will be the data source used for selecting regions
        to project.
    origin : string
        The origin of the coordinate system in the file. If "domain", then the
        center coordinates will be the same as the center of the image as
        defined by the *center* keyword argument. If "image", then the center
        coordinates will be set to (0,0). Default: "domain"
    """

    def __init__(
        self,
        ds,
        axis,
        fields,
        image_res=512,
        center="center",
        width=None,
        depth=(1, "1"),
        weight_field=None,
        length_unit=None,
        deposition="ngp",
        density=False,
        field_parameters=None,
        data_source=None,
        *,
        origin="domain",
    ):
        fields = list(iter_fields(fields))
        axis = fix_axis(axis, ds)
        center, dcenter = ds.coordinates.sanitize_center(center, axis)
        width = ds.coordinates.sanitize_width(axis, width, depth)
        width[-1].convert_to_units(width[0].units)

        if field_parameters is None:
            field_parameters = {}

        ps = ParticleAxisAlignedDummyDataSource(
            center,
            ds,
            axis,
            width,
            fields,
            weight_field=weight_field,
            field_parameters=field_parameters,
            data_source=data_source,
            deposition=deposition,
            density=density,
        )
        w, frb, lunit = construct_image(
            ds, axis, ps, dcenter, image_res, width, length_unit, origin=origin
        )
        super().__init__(frb, fields=fields, length_unit=lunit, wcs=w)


class FITSParticleOffAxisProjection(FITSImageData):
    r"""
    Generate a FITSImageData of an off-axis projection of a
    particle field.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        The dataset object.
    axis : character or integer
        The axis along which to project. One of "x","y","z", or 0,1,2.
    fields : string or list of strings
        The fields to project
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : A sequence of floats, a string, or a tuple.
        The coordinate of the center of the image. If set to 'c', 'center' or
        left blank, the plot is centered on the middle of the domain. If set
        to 'max' or 'm', the center will be located at the maximum of the
        ('gas', 'density') field. Centering on the max or min of a specific
        field is supported by providing a tuple such as ("min","temperature")
        or ("max","dark_matter_density"). Units can be specified by passing in
        *center* as a tuple containing a coordinate and string unit name or by
        passing in a YTArray. If a list or unitless array is supplied, code
        units are assumed.
    width : tuple or a float.
        Width can have four different formats to support variable
        x and y widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs
        wide in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a
        width that is 10 kiloparsecs wide along the x-axis and 15
        kiloparsecs wide along the y-axis.  In the other two examples, code
        units are assumed, for example (0.2, 0.3) specifies a width that has an
        x width of 0.2 and a y width of 0.3 in code units.
    depth : A tuple or a float
         A tuple containing the depth to project through and the string
         key of the unit: (width, 'unit').  If set to a float, code units
         are assumed. Defaults to the entire domain.
    weight_field : string
        The field used to weight the projection.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    deposition : string, optional
        Controls the order of the interpolation of the particles onto the
        mesh. "ngp" is 0th-order "nearest-grid-point" method (the default),
        "cic" is 1st-order "cloud-in-cell".
    density : boolean, optional
        If True, the quantity to be projected will be divided by the area of
        the cells, to make a projected density of the quantity. Default: False
    field_parameters : dictionary
         A dictionary of field parameters than can be accessed by derived
         fields.
    data_source : yt.data_objects.data_containers.YTSelectionContainer, optional
        If specified, this will be the data source used for selecting regions
        to project.
    origin : string
        The origin of the coordinate system in the file. If "domain", then the
        center coordinates will be the same as the center of the image as
        defined by the *center* keyword argument. If "image", then the center
        coordinates will be set to (0,0). Default: "domain"
    """

    def __init__(
        self,
        ds,
        normal,
        fields,
        image_res=512,
        center="c",
        width=None,
        depth=(1, "1"),
        weight_field=None,
        length_unit=None,
        deposition="ngp",
        density=False,
        field_parameters=None,
        data_source=None,
        north_vector=None,
    ):
        fields = list(iter_fields(fields))
        center, dcenter = ds.coordinates.sanitize_center(center, None)
        width = ds.coordinates.sanitize_width(normal, width, depth)
        wd = tuple(el.in_units("code_length").v for el in width)

        if field_parameters is None:
            field_parameters = {}

        ps = ParticleOffAxisDummyDataSource(
            center,
            ds,
            normal,
            wd,
            fields,
            weight_field=weight_field,
            field_parameters=field_parameters,
            data_source=data_source,
            deposition=deposition,
            density=density,
            north_vector=north_vector,
        )
        w, frb, lunit = construct_image(
            ds,
            normal,
            ps,
            dcenter,
            image_res,
            width,
            length_unit,
            origin="image",
        )
        super().__init__(frb, fields=fields, length_unit=lunit, wcs=w)


class FITSOffAxisSlice(FITSImageData):
    r"""
    Generate a FITSImageData of an off-axis slice.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        The dataset object.
    normal : a sequence of floats
        The vector normal to the projection plane.
    fields : string or list of strings
        The fields to slice
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : 'center', 'c', id of a global extremum, or array-like
        The coordinate of the selection's center.
        Defaults to the 'center', i.e. center of the domain.

        Centering on the min or max of a field is supported by passing a tuple
        such as ('min', ('gas', 'density')) or ('max', ('gas', 'temperature'). A
        single string may also be used (e.g. "min_density" or
        "max_temperature"), though it's not as flexible and does not allow to
        select an exact field/particle type. With this syntax, the first field
        matching the provided name is selected.
        'max' or 'm' can be used as a shortcut for ('max', ('gas', 'density'))
        'min' can be used as a shortcut for ('min', ('gas', 'density'))

        One can also select an exact point as a 3 element coordinate sequence,
        e.g. [0.5, 0.5, 0]
        Units can be specified by passing in *center* as a tuple containing a
        3-element coordinate sequence and string unit name, e.g. ([0, 0.5, 0.5], "cm"),
        or by passing in a YTArray. Code units are assumed if unspecified.

    width : tuple or a float.
        Width can have four different formats to support variable
        x and y widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs
        wide in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a
        width that is 10 kiloparsecs wide along the x axis and 15
        kiloparsecs wide along the y axis.  In the other two examples, code
        units are assumed, for example (0.2, 0.3) specifies a width that has an
        x width of 0.2 and a y width of 0.3 in code units.
    north_vector : a sequence of floats
        A vector defining the 'up' direction in the plot.  This
        option sets the orientation of the slicing plane.  If not
        set, an arbitrary grid-aligned north-vector is chosen.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    """

    def __init__(
        self,
        ds,
        normal,
        fields,
        image_res=512,
        center="center",
        width=None,
        north_vector=None,
        length_unit=None,
    ):
        fields = list(iter_fields(fields))
        center, dcenter = ds.coordinates.sanitize_center(center, axis=None)
        cut = ds.cutting(normal, center, north_vector=north_vector)
        center = ds.arr([0.0] * 2, "code_length")
        w, frb, lunit = construct_image(
            ds, normal, cut, center, image_res, width, length_unit
        )
        super().__init__(frb, fields=fields, length_unit=lunit, wcs=w)


class FITSOffAxisProjection(FITSImageData):
    r"""
    Generate a FITSImageData of an off-axis projection.

    Parameters
    ----------
    ds : :class:`~yt.data_objects.static_output.Dataset`
        This is the dataset object corresponding to the
        simulation output to be plotted.
    normal : a sequence of floats
        The vector normal to the projection plane.
    fields : string, list of strings
        The name of the field(s) to be plotted.
    image_res : an int or 2-tuple of ints
        Specify the resolution of the resulting image. A single value will be
        used for both axes, whereas a tuple of values will be used for the
        individual axes. Default: 512
    center : 'center', 'c', id of a global extremum, or array-like
        The coordinate of the selection's center.
        Defaults to the 'center', i.e. center of the domain.

        Centering on the min or max of a field is supported by passing a tuple
        such as ('min', ('gas', 'density')) or ('max', ('gas', 'temperature'). A
        single string may also be used (e.g. "min_density" or
        "max_temperature"), though it's not as flexible and does not allow to
        select an exact field/particle type. With this syntax, the first field
        matching the provided name is selected.
        'max' or 'm' can be used as a shortcut for ('max', ('gas', 'density'))
        'min' can be used as a shortcut for ('min', ('gas', 'density'))

        One can also select an exact point as a 3 element coordinate sequence,
        e.g. [0.5, 0.5, 0]
        Units can be specified by passing in *center* as a tuple containing a
        3-element coordinate sequence and string unit name, e.g. ([0, 0.5, 0.5], "cm"),
        or by passing in a YTArray. Code units are assumed if unspecified.
    width : tuple or a float.
        Width can have four different formats to support variable
        x and y widths.  They are:

        ==================================     =======================
        format                                 example
        ==================================     =======================
        (float, string)                        (10,'kpc')
        ((float, string), (float, string))     ((10,'kpc'),(15,'kpc'))
        float                                  0.2
        (float, float)                         (0.2, 0.3)
        ==================================     =======================

        For example, (10, 'kpc') specifies a width that is 10 kiloparsecs
        wide in the x and y directions, ((10,'kpc'),(15,'kpc')) specifies a
        width that is 10 kiloparsecs wide along the x-axis and 15
        kiloparsecs wide along the y-axis. In the other two examples, code
        units are assumed, for example (0.2, 0.3) specifies a width that has an
        x width of 0.2 and a y width of 0.3 in code units.
    depth : A tuple or a float
        A tuple containing the depth to project through and the string
        key of the unit: (width, 'unit').  If set to a float, code units
        are assumed
    weight_field : string
         The name of the weighting field.  Set to None for no weight.
    north_vector : a sequence of floats
        A vector defining the 'up' direction in the plot.  This
        option sets the orientation of the slicing plane.  If not
        set, an arbitrary grid-aligned north-vector is chosen.
    method : string
        The method of projection.  Valid methods are:

        "integrate" with no weight_field specified : integrate the requested
        field along the line of sight.

        "integrate" with a weight_field specified : weight the requested
        field by the weighting field and integrate along the line of sight.

        "sum" : This method is the same as integrate, except that it does not
        multiply by a path length when performing the integration, and is
        just a straight summation of the field along the given axis. WARNING:
        This should only be used for uniform resolution grid datasets, as other
        datasets may result in unphysical images.
    data_source : yt.data_objects.data_containers.YTSelectionContainer, optional
        If specified, this will be the data source used for selecting regions
        to project.
    length_unit : string, optional
        the length units that the coordinates are written in. The default
        is to use the default length unit of the dataset.
    moment : integer, optional
        for a weighted projection, moment = 1 (the default) corresponds to a
        weighted average. moment = 2 corresponds to a weighted standard
        deviation.
    """

    def __init__(
        self,
        ds,
        normal,
        fields,
        center="center",
        width=(1.0, "unitary"),
        weight_field=None,
        image_res=512,
        data_source=None,
        north_vector=None,
        depth=(1.0, "unitary"),
        method="integrate",
        length_unit=None,
        *,
        moment=1,
    ):
        validate_moment(moment, weight_field)
        center, dcenter = ds.coordinates.sanitize_center(center, axis=None)
        fields = list(iter_fields(fields))
        buf = {}
        width = ds.coordinates.sanitize_width(normal, width, depth)
        wd = tuple(el.in_units("code_length").v for el in width)
        if not is_sequence(image_res):
            image_res = (image_res, image_res)
        res = (image_res[0], image_res[1])
        if data_source is None:
            source = ds.all_data()
        else:
            source = data_source
        fields = source._determine_fields(list(iter_fields(fields)))
        stddev_str = "_stddev" if moment == 2 else ""
        for item in fields:
            ftype, fname = item
            key = (ftype, f"{fname}{stddev_str}")

            buf[key] = off_axis_projection(
                source,
                center,
                normal,
                wd,
                res,
                item,
                north_vector=north_vector,
                method=method,
                weight=weight_field,
            ).swapaxes(0, 1)

            if moment == 2:

                def _sq_field(field, data, item: FieldKey):
                    return data[item] ** 2

                fd = ds._get_field_info(item)
                field_sq = (ftype, f"tmp_{fname}_squared")

                ds.add_field(
                    field_sq,
                    partial(_sq_field, item=item),
                    sampling_type=fd.sampling_type,
                    units=f"({fd.units})*({fd.units})",
                )

                buff2 = off_axis_projection(
                    source,
                    center,
                    normal,
                    wd,
                    res,
                    field_sq,
                    north_vector=north_vector,
                    method=method,
                    weight=weight_field,
                ).swapaxes(0, 1)

                buf[key] = compute_stddev_image(buff2, buf[key])

                ds.field_info.pop(field_sq)

        center = ds.arr([0.0] * 2, "code_length")
        w, not_an_frb, lunit = construct_image(
            ds, normal, buf, center, image_res, width, length_unit
        )
        super().__init__(buf, fields=list(buf.keys()), wcs=w, length_unit=lunit, ds=ds)



from itertools import product

import numpy as np
import pytest

from yt.testing import fake_amr_ds
from yt.visualization.plot_window import ProjectionPlot, SlicePlot


@pytest.fixture(scope="module")
def ds():
    return fake_amr_ds(geometry="cartesian")


@pytest.mark.parametrize("plot_cls", (SlicePlot, ProjectionPlot))
def test_normalplot_all_positional_args(ds, plot_cls):
    plot_cls(ds, "z", ("stream", "Density"))


@pytest.mark.parametrize("plot_cls", (SlicePlot, ProjectionPlot))
def test_normalplot_normal_kwarg(ds, plot_cls):
    plot_cls(ds, normal="z", fields=("stream", "Density"))


@pytest.mark.parametrize("plot_cls", (SlicePlot, ProjectionPlot))
def test_error_with_missing_fields_and_normal(ds, plot_cls):
    with pytest.raises(
        TypeError,
        match="missing 2 required positional arguments: 'normal' and 'fields'",
    ):
        plot_cls(ds)


@pytest.mark.parametrize("plot_cls", (SlicePlot, ProjectionPlot))
def test_error_with_missing_fields_with_normal_kwarg(ds, plot_cls):
    with pytest.raises(
        TypeError, match=r"missing (1 )?required positional argument: 'fields'$"
    ):
        plot_cls(ds, normal="z")


@pytest.mark.parametrize("plot_cls", (SlicePlot, ProjectionPlot))
def test_error_with_missing_fields_with_positional(ds, plot_cls):
    with pytest.raises(
        TypeError, match=r"missing (1 )?required positional argument: 'fields'$"
    ):
        plot_cls(ds, "z")


@pytest.mark.parametrize(
    "plot_cls, normal",
    product([SlicePlot, ProjectionPlot], [(0, 0, 1), [0, 0, 1], np.array((0, 0, 1))]),
)
def test_normalplot_normal_array(ds, plot_cls, normal):
    # see regression https://github.com/yt-project/yt/issues/3736
    plot_cls(ds, normal, fields=("stream", "Density"))



# For this example we will use h5py to write to our output file.
import h5py

import yt

ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

level = 2
dims = ds.domain_dimensions * ds.refine_by**level

# We construct an object that describes the data region and structure we want
# In this case, we want all data up to the maximum "level" of refinement
# across the entire simulation volume.  Higher levels than this will not
# contribute to our covering grid.
cube = ds.covering_grid(
    level,
    left_edge=[0.0, 0.0, 0.0],
    dims=dims,
    # And any fields to preload (this is optional!)
    fields=[("gas", "density")],
)

# Now we open our output file using h5py
# Note that we open with 'w' (write), which will overwrite existing files!
f = h5py.File("my_data.h5", mode="w")

# We create a dataset at the root, calling it "density"
f.create_dataset("/density", data=cube[("gas", "density")])

# We close our file
f.close()

# If we want to then access this datacube in the h5 file, we can now...
f = h5py.File("my_data.h5", mode="r")
print(f["density"][()])



# We don't need to import 'exceptions'
import os.path
from typing import List

from unyt.exceptions import UnitOperationError

from yt._typing import FieldKey


class YTException(Exception):
    pass


# Data access exceptions:


class YTUnidentifiedDataType(YTException):
    def __init__(self, filename, *args, **kwargs):
        self.filename = filename
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        msg = f"Could not determine input format from {self.filename!r}"
        if self.args:
            msg += ", " + (", ".join(f"{a!r}" for a in self.args))
        if self.kwargs:
            msg += ", " + (", ".join(f"{k}={v!r}" for k, v in self.kwargs.items()))
        return msg


class YTAmbiguousDataType(YTUnidentifiedDataType):
    def __init__(self, filename, candidates):
        self.filename = filename
        self.candidates = candidates

    def __str__(self):
        msg = f"Multiple data type candidates for {self.filename}\n"
        msg += "The following independent classes were detected as valid :\n"
        for c in self.candidates:
            msg += f"{c}\n"
        msg += (
            "This degeneracy can be lifted using the `hint` keyword argument in yt.load"
        )
        return msg


class YTSphereTooSmall(YTException):
    def __init__(self, ds, radius, smallest_cell):
        self.ds = ds
        self.radius = radius
        self.smallest_cell = smallest_cell

    def __str__(self):
        return f"{self.radius:0.5e} < {self.smallest_cell:0.5e}"


class YTAxesNotOrthogonalError(YTException):
    def __init__(self, axes):
        self.axes = axes

    def __str__(self):
        return f"The supplied axes are not orthogonal.  {self.axes}"


class YTNoDataInObjectError(YTException):
    def __init__(self, obj):
        self.obj_type = getattr(obj, "_type_name", "")

    def __str__(self):
        s = "The object requested has no data included in it."
        if self.obj_type == "slice":
            s += "  It may lie on a grid face.  Try offsetting slightly."
        return s


class YTFieldNotFound(YTException):
    def __init__(self, field, ds):
        self.field = field
        self.ds = ds

    def _get_suggestions(self) -> List[FieldKey]:
        from yt.funcs import levenshtein_distance

        field = self.field
        ds = self.ds

        suggestions = {}
        if not isinstance(field, tuple):
            ftype, fname = None, field
        elif field[1] is None:
            ftype, fname = None, field[0]
        else:
            ftype, fname = field

        # Limit the suggestions to a distance of 3 (at most 3 edits)
        # This is very arbitrary, but is picked so that...
        # - small typos lead to meaningful suggestions (e.g. `densty` -> `density`)
        # - we don't suggest unrelated things (e.g. `pressure` -> `density` has a distance
        #   of 6, we definitely do not want it)
        # A threshold of 3 seems like a good middle point.
        max_distance = 3

        # Suggest (ftype, fname), with alternative ftype
        for ft, fn in ds.derived_field_list:
            if fn.lower() == fname.lower() and (
                ftype is None or ft.lower() != ftype.lower()
            ):
                suggestions[ft, fn] = 0

        if ftype is not None:
            # Suggest close matches using levenshtein distance
            fields_str = {_: str(_).lower() for _ in ds.derived_field_list}
            field_str = str(field).lower()

            for (ft, fn), fs in fields_str.items():
                distance = levenshtein_distance(field_str, fs, max_dist=max_distance)
                if distance < max_distance:
                    if (ft, fn) in suggestions:
                        continue
                    suggestions[ft, fn] = distance

        # Return suggestions sorted by increasing distance (first are most likely)
        return [
            (ft, fn)
            for (ft, fn), distance in sorted(suggestions.items(), key=lambda v: v[1])
        ]

    def __str__(self):
        msg = f"Could not find field {self.field!r} in {self.ds}."
        try:
            suggestions = self._get_suggestions()
        except AttributeError:
            # This may happen if passing a field that is e.g. an Ellipsis
            # e.g. when using ds.r[...]
            suggestions = []
        if suggestions:
            msg += "\nDid you mean:\n\t"
            msg += "\n\t".join(str(_) for _ in suggestions)
        return msg


class YTParticleTypeNotFound(YTException):
    def __init__(self, fname, ds):
        self.fname = fname
        self.ds = ds

    def __str__(self):
        return f"Could not find particle_type {self.fname!r} in {self.ds}."


class YTSceneFieldNotFound(YTException):
    pass


class YTCouldNotGenerateField(YTFieldNotFound):
    def __str__(self):
        return f"Could field '{self.fname}' in {self.ds} could not be generated."


class YTFieldTypeNotFound(YTException):
    def __init__(self, ftype, ds=None):
        self.ftype = ftype
        self.ds = ds

    def __str__(self):
        if self.ds is not None and self.ftype in self.ds.particle_types:
            return (
                f"Could not find field type {self.ftype!r}. "
                "This field type is a known particle type for this dataset. "
                "Try adding this field with sampling_type='particle'."
            )
        else:
            return f"Could not find field type {self.ftype!r}."


class YTSimulationNotIdentified(YTException):
    def __init__(self, sim_type):
        self.sim_type = sim_type

    def __str__(self):
        return f"Simulation time-series type {self.sim_type!r} not defined."


class YTCannotParseFieldDisplayName(YTException):
    def __init__(self, field_name, display_name, mathtext_error):
        self.field_name = field_name
        self.display_name = display_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return (
            f"The display name {self.display_name!r} of the derived field {self.field_name!r} "
            f"contains the following LaTeX parser errors:\n{self.mathtext_error}"
        )


class YTCannotParseUnitDisplayName(YTException):
    def __init__(self, field_name, unit_name, mathtext_error):
        self.field_name = field_name
        self.unit_name = unit_name
        self.mathtext_error = mathtext_error

    def __str__(self):
        return (
            f"The unit display name {self.unit_name!r} of the derived field {self.field_name!r} "
            f"contains the following LaTeX parser errors:\n{self.mathtext_error}"
        )


class InvalidSimulationTimeSeries(YTException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class MissingParameter(YTException):
    def __init__(self, ds, parameter):
        self.ds = ds
        self.parameter = parameter

    def __str__(self):
        return f"dataset {self.ds} is missing {self.parameter} parameter."


class NoStoppingCondition(YTException):
    def __init__(self, ds):
        self.ds = ds

    def __str__(self):
        return f"Simulation {self.ds} has no stopping condition. StopTime or StopCycle should be set."


class YTNotInsideNotebook(YTException):
    def __str__(self):
        return "This function only works from within an IPython Notebook."


class YTCoordinateNotImplemented(YTException):
    def __str__(self):
        return "This coordinate is not implemented for this geometry type."


# define for back compat reasons for code written before yt 4.0
YTUnitOperationError = UnitOperationError


class YTUnitNotRecognized(YTException):
    def __init__(self, unit):
        self.unit = unit

    def __str__(self):
        return f"This dataset doesn't recognize {self.unit!r}"


class YTFieldUnitError(YTException):
    def __init__(self, field_info, returned_units):
        self.msg = (
            f"The field function associated with the field {field_info.name!r} returned "
            f"data with units {returned_units!r} but was defined with units {field_info.units!r}."
        )

    def __str__(self):
        return self.msg


class YTFieldUnitParseError(YTException):
    def __init__(self, field_info):
        self.msg = (
            f"The field {field_info.name!r} has unparsable units {field_info.units!r}."
        )

    def __str__(self):
        return self.msg


class YTSpatialFieldUnitError(YTException):
    def __init__(self, field):
        self.msg = (
            f"Field {field!r} is a spatial field but has unknown units but "
            "spatial fields must have explicitly defined units. Add the "
            "field with explicit 'units' to clear this error."
        )

    def __str__(self):
        return self.msg


class YTHubRegisterError(YTException):
    def __str__(self):
        return (
            "You must create an API key before uploading. See "
            "https://data.yt-project.org/getting_started.html"
        )


class YTNoFilenamesMatchPattern(YTException):
    def __init__(self, pattern):
        self.pattern = pattern

    def __str__(self):
        return f"No filenames were found to match the pattern: {self.pattern!r}"


class YTNoOldAnswer(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return f"There is no old answer available.\n{self.path!r}"


class YTNoAnswerNameSpecified(YTException):
    def __init__(self, message=None):
        if message is None or message == "":
            message = (
                "Answer name not provided for the answer testing test."
                "\n  Please specify --answer-name=<answer_name> in"
                " command line mode or in AnswerTestingTest.answer_name"
                " variable."
            )
        self.message = message

    def __str__(self):
        return str(self.message)


class YTCloudError(YTException):
    def __init__(self, path):
        self.path = path

    def __str__(self):
        return (
            f"Failed to retrieve cloud data. Connection may be broken.\n {self.path!r}"
        )


class YTEllipsoidOrdering(YTException):
    def __init__(self, ds, A, B, C):
        self.ds = ds
        self._A = A
        self._B = B
        self._C = C

    def __str__(self):
        return "Must have A>=B>=C"


class EnzoTestOutputFileNonExistent(YTException):
    def __init__(self, filename):
        self.filename = filename
        self.testname = os.path.basename(os.path.dirname(filename))

    def __str__(self):
        return (
            f"Enzo test output file (OutputLog) not generated for: {self.testname!r}.\n"
            "Test did not complete."
        )


class YTNoAPIKey(YTException):
    def __init__(self, service, config_name):
        self.service = service
        self.config_name = config_name

    def __str__(self):
        from yt.config import config_dir

        try:
            conf = os.path.join(config_dir(), "yt", "yt.toml")
        except Exception:
            # this is really not a good time to raise another exception
            conf = "yt's configuration file"
        return f"You need to set an API key for {self.service!r} in {conf} as {self.config_name!r}"


class YTTooManyVertices(YTException):
    def __init__(self, nv, fn):
        self.nv = nv
        self.fn = fn

    def __str__(self):
        s = f"There are too many vertices ({self.nv}) to upload to Sketchfab. "
        s += f"Your model has been saved as {self.fn} .  You should upload manually."
        return s


class YTInvalidWidthError(YTException):
    def __init__(self, width):
        self.error = f"width ({str(width)}) is invalid"

    def __str__(self):
        return str(self.error)


class YTFieldNotParseable(YTException):
    def __init__(self, field):
        self.field = field

    def __str__(self):
        return f"Cannot identify field {self.field!r}"


class YTDataSelectorNotImplemented(YTException):
    def __init__(self, class_name):
        self.class_name = class_name

    def __str__(self):
        return f"Data selector {self.class_name!r} not implemented."


class YTParticleDepositionNotImplemented(YTException):
    def __init__(self, class_name):
        self.class_name = class_name

    def __str__(self):
        return f"Particle deposition method {self.class_name!r} not implemented."


class YTDomainOverflow(YTException):
    def __init__(self, mi, ma, dle, dre):
        self.mi = mi
        self.ma = ma
        self.dle = dle
        self.dre = dre

    def __str__(self):
        return (
            f"Particle bounds {self.mi} and {self.ma} "
            f"exceed domain bounds {self.dle} and {self.dre}"
        )


class YTIntDomainOverflow(YTException):
    def __init__(self, dims, dd):
        self.dims = dims
        self.dd = dd

    def __str__(self):
        return f"Integer domain overflow: {self.dims} in {self.dd}"


class YTIllDefinedFilter(YTException):
    def __init__(self, filter, s1, s2):
        self.filter = filter
        self.s1 = s1
        self.s2 = s2

    def __str__(self):
        return (
            f"Filter {self.filter!r} ill-defined. "
            f"Applied to shape {self.s1} but is shape {self.s2}."
        )


class YTIllDefinedParticleFilter(YTException):
    def __init__(self, filter, missing):
        self.filter = filter
        self.missing = missing

    def __str__(self):
        msg = (
            '\nThe fields\n\t{},\nrequired by the "{}" particle filter, '
            "are not defined for this dataset."
        )
        f = self.filter
        return msg.format("\n".join(str(m) for m in self.missing), f.name)


class YTIllDefinedBounds(YTException):
    def __init__(self, lb, ub):
        self.lb = lb
        self.ub = ub

    def __str__(self):
        v = f"The bounds {self.lb:0.3e} and {self.ub:0.3e} are ill-defined. "
        v += "Typically this happens when a log binning is specified "
        v += "and zero or negative values are given for the bounds."
        return v


class YTObjectNotImplemented(YTException):
    def __init__(self, ds, obj_name):
        self.ds = ds
        self.obj_name = obj_name

    def __str__(self):
        return f"The object type {self.obj_name!r} is not implemented for the dataset {self.ds!s}"


class YTParticleOutputFormatNotImplemented(YTException):
    def __str__(self):
        return "The particle output format is not supported."


class YTFileNotParseable(YTException):
    def __init__(self, fname, line):
        self.fname = fname
        self.line = line

    def __str__(self):
        return f"Error while parsing file {self.fname!r} at line {self.line}"


class YTRockstarMultiMassNotSupported(YTException):
    def __init__(self, mi, ma, ptype):
        self.mi = mi
        self.ma = ma
        self.ptype = ptype

    def __str__(self):
        v = f"Particle type '{self.ptype}' has minimum mass {self.mi:0.3e} and maximum "
        v += f"mass {self.ma:0.3e}.  Multi-mass particles are not currently supported."
        return v


class YTTooParallel(YTException):
    def __str__(self):
        return "You've used too many processors for this dataset."


class YTElementTypeNotRecognized(YTException):
    def __init__(self, dim, num_nodes):
        self.dim = dim
        self.num_nodes = num_nodes

    def __str__(self):
        return f"Element type not recognized - dim = {self.dim}, num_nodes = {self.num_nodes}"


class YTDuplicateFieldInProfile(YTException):
    def __init__(self, field, new_spec, old_spec):
        self.field = field
        self.new_spec = new_spec
        self.old_spec = old_spec

    def __str__(self):
        r = f"""Field {self.field} already exists with field spec:
               {self.old_spec}
               But being asked to add it with:
               {self.new_spec}"""
        return r


class YTInvalidPositionArray(YTException):
    def __init__(self, shape, dimensions):
        self.shape = shape
        self.dimensions = dimensions

    def __str__(self):
        r = f"""Position arrays must be length and shape (N,3).
               But this one has {self.dimensions} and {self.shape}."""
        return r


class YTIllDefinedCutRegion(YTException):
    def __init__(self, conditions):
        self.conditions = conditions

    def __str__(self):
        r = (
            "Can't mix particle/discrete and fluid/mesh conditions or quantities. "
            "Conditions specified:\n"
        )
        r += "\n".join(c for c in self.conditions)
        return r


class YTMixedCutRegion(YTException):
    def __init__(self, conditions, field):
        self.conditions = conditions
        self.field = field

    def __str__(self):
        r = f"""Can't mix particle/discrete and fluid/mesh conditions or
               quantities.  Field: {self.field} and Conditions specified:
            """
        r += "\n".join(c for c in self.conditions)
        return r


class YTGDFAlreadyExists(YTException):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return f"A file already exists at {self.filename} and overwrite=False."


class YTNonIndexedDataContainer(YTException):
    def __init__(self, cont):
        self.cont = cont

    def __str__(self):
        class_name = self.cont.__class__.__name__
        return (
            f"The data container type ({class_name}) is an unindexed type. "
            "Operations such as ires, icoords, fcoords and fwidth will not work on it.\n"
            "Did you just attempt to perform an off-axis operation ? "
            "Be sure to consult the latest documentation to see whether the operation "
            "you tried is actually supported for your data type."
        )


class YTGDFUnknownGeometry(YTException):
    def __init__(self, geometry):
        self.geometry = geometry

    def __str__(self):
        return (
            """Unknown geometry %i. Please refer to GDF standard
                  for more information"""
            % self.geometry
        )


class YTInvalidUnitEquivalence(YTException):
    def __init__(self, equiv, unit1, unit2):
        self.equiv = equiv
        self.unit1 = unit1
        self.unit2 = unit2

    def __str__(self):
        return f"The unit equivalence {self.equiv!r} does not exist for the units {self.unit1!r} and {self.unit2!r}."


class YTPlotCallbackError(YTException):
    def __init__(self, callback):
        self.callback = "annotate_" + callback

    def __str__(self):
        return f"{self.callback} callback failed"


class YTUnsupportedPlotCallback(YTPlotCallbackError):
    def __init__(self, callback: str, plot_type: str) -> None:
        super().__init__(callback)
        self.plot_type = plot_type

    def __str__(self):
        return f"The `{self.plot_type}` class currently doesn't support the `{self.callback}` method."


class YTPixelizeError(YTException):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class YTDimensionalityError(YTException):
    def __init__(self, wrong, right):
        self.wrong = wrong
        self.right = right

    def __str__(self):
        return f"Dimensionality specified was {self.wrong} but we need {self.right}"


class YTInvalidShaderType(YTException):
    def __init__(self, source):
        self.source = source

    def __str__(self):
        return f"Can't identify shader_type for file {self.source!r}"


class YTInvalidFieldType(YTException):
    def __init__(self, fields):
        self.fields = fields

    def __str__(self):
        return (
            "\nSlicePlot, ProjectionPlot, and OffAxisProjectionPlot can "
            "only plot fields that\n"
            "are defined on a mesh or for SPH particles, but received the "
            "following N-body\n"
            "particle fields:\n\n"
            f"    {self.fields!r}\n\n"
            "Did you mean to use ParticlePlot or plot a deposited particle "
            "field instead?"
        )


class YTUnknownUniformKind(YTException):
    def __init__(self, kind):
        self.kind = kind

    def __str__(self):
        return f"Can't determine kind specification for {self.kind!r}"


class YTUnknownUniformSize(YTException):
    def __init__(self, size_spec):
        self.size_spec = size_spec

    def __str__(self):
        return f"Can't determine size specification for {self.size_spec!r}"


class YTDataTypeUnsupported(YTException):
    def __init__(self, this, supported):
        self.supported = supported
        self.this = this

    def __str__(self):
        v = f"This operation is not supported for data of geometry {self.this!r}; "
        v += f"It supports data of geometries {self.supported!r}"
        return v


class YTBoundsDefinitionError(YTException):
    def __init__(self, message, bounds):
        self.bounds = bounds
        self.message = message

    def __str__(self):
        v = f"This operation has encountered a bounds error: {self.message} "
        v += f"\nSpecified bounds are {self.bounds!r}."
        return v


def screen_one_element_list(lis):
    if len(lis) == 1:
        return lis[0]
    return lis


class YTIllDefinedProfile(YTException):
    def __init__(self, bin_fields, fields, weight_field, is_pfield):
        nbin = len(bin_fields)
        nfields = len(fields)
        self.bin_fields = screen_one_element_list(bin_fields)
        self.bin_fields_ptype = screen_one_element_list(is_pfield[:nbin])
        self.fields = screen_one_element_list(fields)
        self.fields_ptype = screen_one_element_list(is_pfield[nbin : nbin + nfields])
        self.weight_field = weight_field
        if self.weight_field is not None:
            self.weight_field_ptype = is_pfield[-1]

    def __str__(self):
        msg = (
            "\nCannot create a profile object that mixes particle and mesh "
            "fields.\n\n"
            "Received the following bin_fields:\n\n"
            "   %s, particle_type = %s\n\n"
            "Profile fields:\n\n"
            "   %s, particle_type = %s\n"
        )
        msg = msg % (
            self.bin_fields,
            self.bin_fields_ptype,
            self.fields,
            self.fields_ptype,
        )

        if self.weight_field is not None:
            weight_msg = "\nAnd weight field:\n\n   %s, particle_type = %s\n"
            weight_msg = weight_msg % (self.weight_field, self.weight_field_ptype)
        else:
            weight_msg = ""

        return msg + weight_msg


class YTProfileDataShape(YTException):
    def __init__(self, field1, shape1, field2, shape2):
        self.field1 = field1
        self.shape1 = shape1
        self.field2 = field2
        self.shape2 = shape2

    def __str__(self):
        return (
            "Profile fields must have same shape: {self.field1!r} has "
            f"shape {self.shape1} and {self.field2!r} has shape {self.shape2}."
        )


class YTBooleanObjectError(YTException):
    def __init__(self, bad_object):
        self.bad_object = bad_object

    def __str__(self):
        v = f"Supplied:\n{self.bad_object}\nto a boolean operation"
        v += " but it is not a YTSelectionContainer3D object."
        return v


class YTBooleanObjectsWrongDataset(YTException):
    def __init__(self):
        pass

    def __str__(self):
        return "Boolean data objects must share a common dataset object."


class YTIllDefinedAMR(YTException):
    def __init__(self, level, axis):
        self.level = level
        self.axis = axis

    def __str__(self):
        return (
            f"Grids on the level {self.level} are not properly aligned with cell edges "
            f"on the parent level ({self.axis!r} axis)"
        )


class YTIllDefinedParticleData(YTException):
    pass


class YTIllDefinedAMRData(YTException):
    pass


class YTInconsistentGridFieldShape(YTException):
    def __init__(self, shapes):
        self.shapes = shapes

    def __str__(self):
        msg = "Not all grid-based fields have the same shape!\n"
        for name, shape in self.shapes:
            msg += f"    Field {name!r} has shape {shape}.\n"
        return msg


class YTInconsistentParticleFieldShape(YTException):
    def __init__(self, ptype, shapes):
        self.ptype = ptype
        self.shapes = shapes

    def __str__(self):
        msg = "Not all fields with field type {self.ptype!r} have the same shape!\n"
        for name, shape in self.shapes:
            field = (self.ptype, name)
            msg += f"    Field {field} has shape {shape}.\n"
        return msg


class YTInconsistentGridFieldShapeGridDims(YTException):
    def __init__(self, shapes, grid_dims):
        self.shapes = shapes
        self.grid_dims = grid_dims

    def __str__(self):
        msg = "Not all grid-based fields match the grid dimensions! "
        msg += f"Grid dims are {self.grid_dims}, "
        msg += "and the following fields have shapes that do not match them:\n"
        for name, shape in self.shapes:
            if shape != self.grid_dims:
                msg += f"    Field {name} has shape {shape}.\n"
        return msg


class YTCommandRequiresModule(YTException):
    def __init__(self, module: str):
        self.module = module

    def __str__(self):
        msg = f"This command requires {self.module!r} to be installed.\n\n"
        msg += f"Please install {self.module!r} with the package manager "
        msg += "appropriate for your python environment, e.g.:\n"
        msg += f"  conda install {self.module}\n"
        msg += "or:\n"
        msg += f" python -m pip install {self.module}\n"
        return msg


class YTModuleRemoved(YTException):
    def __init__(self, name, new_home=None, info=None):
        message = f"The {name} module has been removed from yt."
        if new_home is not None:
            message += f"\nIt has been moved to {new_home}."
        if info is not None:
            message += f"\nFor more information, see {info}."
        super().__init__(message)


class YTArrayTooLargeToDisplay(YTException):
    def __init__(self, size, max_size):
        self.size = size
        self.max_size = max_size

    def __str__(self):
        msg = f"The requested array is of size {self.size}.\n"
        msg += "We do not support displaying arrays larger\n"
        msg += f"than size {self.max_size}."
        return msg


class YTConfigurationError(YTException):
    pass


class GenerationInProgress(Exception):
    def __init__(self, fields):
        self.fields = fields


class MountError(Exception):
    def __init__(self, message):
        self.message = message



import numpy as np

from yt.utilities.exceptions import YTDomainOverflow
from yt.utilities.io_handler import BaseIOHandler, BaseParticleIOHandler
from yt.utilities.logger import ytLogger as mylog


class IOHandlerStream(BaseIOHandler):
    _dataset_type = "stream"
    _vector_fields = {"particle_velocity": 3, "particle_position": 3}

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        self.field_units = ds.stream_handler.field_units
        super().__init__(ds)

    def _read_data_set(self, grid, field):
        # This is where we implement processor-locking
        tr = self.fields[grid.id][field]
        if callable(tr):
            tr = tr(grid, field)
        # If it's particles, we copy.
        if len(tr.shape) == 1:
            return tr.copy()
        # New in-place unit conversion breaks if we don't copy first
        return tr

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        if any((ftype not in self.ds.fluid_types for ftype, fname in fields)):
            raise NotImplementedError
        rv = {}
        for field in fields:
            rv[field] = self.ds.arr(np.empty(size, dtype="float64"))

        ng = sum(len(c.objs) for c in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s blocks",
            size,
            [f2 for f1, f2 in fields],
            ng,
        )
        for field in fields:
            ftype, fname = field
            ind = 0
            for chunk in chunks:
                for g in chunk.objs:
                    ds = self.fields[g.id][ftype, fname]
                    if callable(ds):
                        ds = ds(g, field)
                    ind += g.select(selector, ds, rv[field], ind)  # caches
        return rv

    def _read_particle_coords(self, chunks, ptf):
        chunks = list(chunks)
        for chunk in chunks:
            for g in chunk.objs:
                if g.NumberOfParticles == 0:
                    continue
                gf = self.fields[g.id]
                for ptype in sorted(ptf):
                    if (ptype, "particle_position") in gf:
                        x, y, z = gf[ptype, "particle_position"].T
                    else:
                        x, y, z = (
                            gf[ptype, f"particle_position_{ax}"]
                            for ax in self.ds.coordinates.axis_order
                        )
                    yield ptype, (x, y, z), 0.0

    def _read_particle_fields(self, chunks, ptf, selector):
        chunks = list(chunks)
        for chunk in chunks:
            for g in chunk.objs:
                if g.NumberOfParticles == 0:
                    continue
                gf = self.fields[g.id]
                for ptype, field_list in sorted(ptf.items()):
                    if (ptype, "particle_position") in gf:
                        x, y, z = gf[ptype, "particle_position"].T
                    else:
                        x, y, z = (
                            gf[ptype, f"particle_position_{ax}"]
                            for ax in self.ds.coordinates.axis_order
                        )
                    mask = selector.select_points(x, y, z, 0.0)
                    if mask is None:
                        continue
                    for field in field_list:
                        data = np.asarray(gf[ptype, field])
                        yield (ptype, field), data[mask]

    @property
    def _read_exception(self):
        return KeyError


class StreamParticleIOHandler(BaseParticleIOHandler):
    _dataset_type = "stream_particles"
    _vector_fields = {"particle_velocity": 3, "particle_position": 3}

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super().__init__(ds)

    def _read_particle_coords(self, chunks, ptf):
        for data_file in sorted(
            self._get_data_files(chunks), key=lambda x: (x.filename, x.start)
        ):
            f = self.fields[data_file.filename]
            # This double-reads
            for ptype in sorted(ptf):
                yield ptype, (
                    f[ptype, "particle_position_x"],
                    f[ptype, "particle_position_y"],
                    f[ptype, "particle_position_z"],
                ), 0.0

    def _read_smoothing_length(self, chunks, ptf, ptype):
        for data_file in sorted(
            self._get_data_files(chunks), key=lambda x: (x.filename, x.start)
        ):
            f = self.fields[data_file.filename]
            return f[ptype, "smoothing_length"]

    def _get_data_files(self, chunks):
        data_files = set()
        for chunk in chunks:
            for obj in chunk.objs:
                data_files.update(obj.data_files)
        return data_files

    def _read_particle_data_file(self, data_file, ptf, selector=None):
        return_data = {}
        f = self.fields[data_file.filename]
        for ptype, field_list in sorted(ptf.items()):
            if (ptype, "particle_position") in f:
                ppos = f[ptype, "particle_position"]
                x = ppos[:, 0]
                y = ppos[:, 1]
                z = ppos[:, 2]
            else:
                x, y, z = (f[ptype, f"particle_position_{ax}"] for ax in "xyz")
            if (ptype, "smoothing_length") in self.ds.field_list:
                hsml = f[ptype, "smoothing_length"]
            else:
                hsml = 0.0

            if selector:
                mask = selector.select_points(x, y, z, hsml)
            if mask is None:
                continue
            for field in field_list:
                data = f[ptype, field]
                if selector:
                    data = data[mask]

                return_data[(ptype, field)] = data

        return return_data

    def _yield_coordinates(self, data_file, needed_ptype=None):
        # self.fields[g.id][fname] is the pattern here
        for ptype in self.ds.particle_types_raw:
            if needed_ptype is not None and needed_ptype is not ptype:
                continue
            try:
                pos = np.column_stack(
                    [
                        self.fields[data_file.filename][
                            (ptype, f"particle_position_{ax}")
                        ]
                        for ax in "xyz"
                    ]
                )
            except KeyError:
                pos = self.fields[data_file.filename][ptype, "particle_position"]
            if np.any(pos.min(axis=0) < data_file.ds.domain_left_edge) or np.any(
                pos.max(axis=0) > data_file.ds.domain_right_edge
            ):
                raise YTDomainOverflow(
                    pos.min(axis=0),
                    pos.max(axis=0),
                    data_file.ds.domain_left_edge,
                    data_file.ds.domain_right_edge,
                )
            yield ptype, pos

    def _get_smoothing_length(self, data_file, dtype, shape):
        ptype = self.ds._sph_ptypes[0]
        return self.fields[data_file.filename][ptype, "smoothing_length"]

    def _count_particles(self, data_file):
        pcount = {}
        for ptype in self.ds.particle_types_raw:
            pcount[ptype] = 0
        # stream datasets only have one "file"
        if data_file.file_id > 0:
            return pcount
        for ptype in self.ds.particle_types_raw:
            d = self.fields[data_file.filename]
            try:
                pcount[ptype] = d[ptype, "particle_position_x"].size
            except KeyError:
                pcount[ptype] = d[ptype, "particle_position"].shape[0]
        return pcount

    def _identify_fields(self, data_file):
        return self.fields[data_file.filename].keys(), {}


class IOHandlerStreamHexahedral(BaseIOHandler):
    _dataset_type = "stream_hexahedral"
    _vector_fields = {"particle_velocity": 3, "particle_position": 3}

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super().__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        assert len(chunks) == 1
        chunk = chunks[0]
        rv = {}
        for field in fields:
            ftype, fname = field
            rv[field] = np.empty(size, dtype="float64")
        ngrids = sum(len(chunk.objs) for chunk in chunks)
        mylog.debug(
            "Reading %s cells of %s fields in %s blocks",
            size,
            [fn for ft, fn in fields],
            ngrids,
        )
        for field in fields:
            ind = 0
            ftype, fname = field
            for chunk in chunks:
                for g in chunk.objs:
                    ds = self.fields[g.mesh_id].get(field, None)
                    if ds is None:
                        ds = self.fields[g.mesh_id][fname]
                    ind += g.select(selector, ds, rv[field], ind)  # caches
        return rv


class IOHandlerStreamOctree(BaseIOHandler):
    _dataset_type = "stream_octree"
    _vector_fields = {"particle_velocity": 3, "particle_position": 3}

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super().__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        rv = {}
        ind = 0
        chunks = list(chunks)
        assert len(chunks) == 1
        for chunk in chunks:
            assert len(chunk.objs) == 1
            for subset in chunk.objs:
                field_vals = {}
                for field in fields:
                    field_vals[field] = self.fields[
                        subset.domain_id - subset._domain_offset
                    ][field]
                subset.fill(field_vals, rv, selector, ind)
        return rv


class IOHandlerStreamUnstructured(BaseIOHandler):
    _dataset_type = "stream_unstructured"

    def __init__(self, ds):
        self.fields = ds.stream_handler.fields
        super().__init__(ds)

    def _read_fluid_selection(self, chunks, selector, fields, size):
        chunks = list(chunks)
        rv = {}
        for field in fields:
            ftype, fname = field
            if ftype == "all":
                ci = np.concatenate(
                    [mesh.connectivity_indices for mesh in self.ds.index.mesh_union]
                )
            else:
                mesh_id = int(ftype[-1]) - 1
                m = self.ds.index.meshes[mesh_id]
                ci = m.connectivity_indices
            num_elem = ci.shape[0]
            if fname in self.ds._node_fields:
                nodes_per_element = ci.shape[1]
                rv[field] = np.empty((num_elem, nodes_per_element), dtype="float64")
            else:
                rv[field] = np.empty(num_elem, dtype="float64")
        for field in fields:
            ind = 0
            ftype, fname = field
            if ftype == "all":
                objs = list(self.ds.index.mesh_union)
            else:
                mesh_ids = [int(ftype[-1])]
                chunk = chunks[mesh_ids[0] - 1]
                objs = chunk.objs
            for g in objs:
                ds = self.fields[g.mesh_id].get(field, None)
                if ds is None:
                    f = ("connect%d" % (g.mesh_id + 1), fname)
                    ds = self.fields[g.mesh_id][f]
                ind += g.select(selector, ds, rv[field], ind)  # caches
            rv[field] = rv[field][:ind]
        return rv



import re

from .periodic_table import periodic_table
from .physical_ratios import _primordial_mass_fraction


class ChemicalFormula:
    def __init__(self, formula_string):
        # See YTEP-0003 for information on the format.
        self.formula_string = formula_string
        self.elements = []
        if "_" in self.formula_string:
            molecule, ionization = self.formula_string.split("_")
            if ionization[0] == "p":
                charge = int(ionization[1:])
            elif ionization[0] == "m":
                charge = -int(ionization[1:])
            else:
                raise NotImplementedError
        elif self.formula_string.startswith("El"):
            molecule = self.formula_string
            charge = -1
        else:
            molecule = self.formula_string
            charge = 0
        self.charge = charge
        for element, count in re.findall(r"([A-Z][a-z]*)(\d*)", molecule):
            if count == "":
                count = 1
            self.elements.append((periodic_table[element], int(count)))
        self.weight = sum(n * e.weight for e, n in self.elements)

    def __repr__(self):
        return self.formula_string


def compute_mu(ion_state):
    if ion_state == "ionized" or ion_state is None:
        # full ionization
        n_H = 2.0  # fully ionized hydrogen gives two particles
        n_He = 3.0  # fully ionized helium gives three particles
    elif ion_state == "neutral":
        n_H = 1.0  # neutral hydrogen gives one particle
        n_He = 1.0  # neutral helium gives one particle
    muinv = n_H * _primordial_mass_fraction["H"] / ChemicalFormula("H").weight
    muinv += n_He * _primordial_mass_fraction["He"] / ChemicalFormula("He").weight
    return 1.0 / muinv



import os
import tempfile

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pytest

from yt.testing import fake_vr_orientation_test_ds
from yt.visualization.volume_rendering.api import (
    Scene,
    create_volume_source,
    off_axis_projection,
)


def scene_to_mpl_figure(scene):
    """helper function to convert a scene image rendering to matplotlib
    so we can rely on pytest-mpl to compare images
    """
    tmpfd, tmpname = tempfile.mkstemp(suffix=".png")
    os.close(tmpfd)
    scene.save(tmpname, sigma_clip=1.0)
    image = mpl.image.imread(tmpname)
    os.remove(tmpname)

    fig, ax = plt.subplots()
    ax.set(aspect="equal")
    ax.imshow(image)
    return fig


class TestOrientation:
    @classmethod
    def setup_class(cls):
        cls.ds = fake_vr_orientation_test_ds()

        cls.scene = Scene()

        vol = create_volume_source(cls.ds, field=("gas", "density"))
        cls.scene.add_source(vol)

        cls._last_lense_type = None

    @classmethod
    def set_camera(cls, lens_type):
        # this method isn't thread-safe
        # if lens_type == cls._last_lense_type:
        #    return

        cls._last_lense_type = lens_type

        cam = cls.scene.add_camera(cls.ds, lens_type=lens_type)
        cam.resolution = (1000, 1000)
        cam.position = cls.ds.arr(np.array([-4.0, 0.0, 0.0]), "code_length")
        cam.switch_orientation(
            normal_vector=[1.0, 0.0, 0.0], north_vector=[0.0, 0.0, 1.0]
        )
        cam.set_width(cls.ds.domain_width * 2.0)
        cls.camera = cam

    @pytest.mark.parametrize("lens_type", ["perspective", "plane-parallel"])
    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_vr_orientation_lense_type(self, lens_type):
        # note that a previous version of this test proved flaky
        # and required a much lower precision for plane-parallel
        # https://github.com/yt-project/yt/issue/3069
        # https://github.com/yt-project/yt/pull/3068
        # https://github.com/yt-project/yt/pull/3294
        self.set_camera(lens_type)
        return scene_to_mpl_figure(self.scene)

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_vr_orientation_yaw(self):
        self.set_camera("plane-parallel")
        center = self.ds.arr([0, 0, 0], "code_length")
        self.camera.yaw(np.pi, rot_center=center)
        return scene_to_mpl_figure(self.scene)

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_vr_orientation_pitch(self):
        self.set_camera("plane-parallel")
        center = self.ds.arr([0, 0, 0], "code_length")
        self.camera.pitch(np.pi, rot_center=center)
        return scene_to_mpl_figure(self.scene)

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_vr_orientation_roll(self):
        self.set_camera("plane-parallel")
        center = self.ds.arr([0, 0, 0], "code_length")
        self.camera.roll(np.pi, rot_center=center)
        return scene_to_mpl_figure(self.scene)

    @pytest.mark.mpl_image_compare(remove_text=True)
    def test_vr_orientation_off_axis_projection(self):
        image = off_axis_projection(
            self.ds,
            center=[0.5, 0.5, 0.5],
            normal_vector=[-0.3, -0.1, 0.8],
            width=[1.0, 1.0, 1.0],
            resolution=512,
            item=("gas", "density"),
            no_ghost=False,
        )

        fig, ax = plt.subplots()
        ax.imshow(image)
        return fig



import inspect
from textwrap import TextWrapper

import yt

ds = yt.load("RD0005-mine/RedshiftOutput0005")

output = open("source/analyzing/_obj_docstrings.inc", "w")

template = """

.. class:: %(clsname)s%(sig)s:

   For more information, see :ref:`%(docstring)s`
   (This is a proxy for :class:`~%(clsproxy)sBase`.)
"""

tw = TextWrapper(initial_indent="   ", subsequent_indent="   ", width=60)


def write_docstring(f, name, cls):
    for clsi in inspect.getmro(cls):
        docstring = inspect.getdoc(clsi.__init__)
        if docstring is not None:
            break
    clsname = name
    sig = inspect.formatargspec(*inspect.getargspec(cls.__init__))
    sig = sig.replace("**kwargs", "**field_parameters")
    clsproxy = f"yt.data_objects.data_containers.{cls.__name__}"
    f.write(
        template
        % dict(
            clsname=clsname, sig=sig, clsproxy=clsproxy, docstring="physical-object-api"
        )
    )


for n, c in sorted(ds.__dict__.items()):
    if hasattr(c, "_con_args"):
        print(n)
        write_docstring(output, n, c)



import numpy as np
from numpy.testing import assert_equal

from yt.frontends.athena.api import AthenaDataset
from yt.loaders import load
from yt.testing import (
    assert_allclose_units,
    disable_dataset_cache,
    requires_file,
)
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields_cloud = (
    ("athena", "scalar[0]"),
    ("gas", "density"),
    ("gas", "total_energy_density"),
)

cloud = "ShockCloud/id0/Cloud.0050.vtk"


@requires_ds(cloud)
def test_cloud():
    ds = data_dir_load(cloud)
    assert_equal(str(ds), "Cloud.0050")
    for test in small_patch_amr(ds, _fields_cloud):
        test_cloud.__name__ = test.description
        yield test


_fields_blast = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("gas", "velocity_magnitude"),
)

blast = "MHDBlast/id0/Blast.0100.vtk"


@requires_ds(blast)
def test_blast():
    ds = data_dir_load(blast)
    assert_equal(str(ds), "Blast.0100")
    for test in small_patch_amr(ds, _fields_blast):
        test_blast.__name__ = test.description
        yield test


uo_blast = {
    "length_unit": (1.0, "pc"),
    "mass_unit": (2.38858753789e-24, "g/cm**3*pc**3"),
    "time_unit": (1.0, "s*pc/km"),
}


@requires_file(blast)
def test_blast_override():
    # verify that overriding units causes derived unit values to be updated.
    # see issue #1259
    ds = load(blast, units_override=uo_blast)
    assert_equal(float(ds.magnetic_unit.in_units("gauss")), 5.47867467969813e-07)


uo_stripping = {
    "time_unit": 3.086e14,
    "length_unit": 8.0236e22,
    "mass_unit": 9.999e-30 * 8.0236e22**3,
}

_fields_stripping = (
    ("gas", "temperature"),
    ("gas", "density"),
    ("athena", "specific_scalar[0]"),
)

stripping = "RamPressureStripping/id0/rps.0062.vtk"


@requires_ds(stripping, big_data=True)
def test_stripping():
    ds = data_dir_load(stripping, kwargs={"units_override": uo_stripping})
    assert_equal(str(ds), "rps.0062")
    for test in small_patch_amr(ds, _fields_stripping):
        test_stripping.__name__ = test.description
        yield test


sloshing = "MHDSloshing/virgo_low_res.0054.vtk"

uo_sloshing = {
    "length_unit": (1.0, "Mpc"),
    "time_unit": (1.0, "Myr"),
    "mass_unit": (1.0e14, "Msun"),
}


@requires_file(sloshing)
@disable_dataset_cache
def test_nprocs():
    ds1 = load(sloshing, units_override=uo_sloshing)
    sp1 = ds1.sphere("c", (100.0, "kpc"))
    prj1 = ds1.proj(("gas", "density"), 0)
    ds2 = load(sloshing, units_override=uo_sloshing, nprocs=8)
    sp2 = ds2.sphere("c", (100.0, "kpc"))
    prj2 = ds1.proj(("gas", "density"), 0)

    assert_equal(
        sp1.quantities.extrema(("gas", "pressure")),
        sp2.quantities.extrema(("gas", "pressure")),
    )
    assert_allclose_units(
        sp1.quantities.total_quantity(("gas", "pressure")),
        sp2.quantities.total_quantity(("gas", "pressure")),
    )
    for ax in "xyz":
        assert_equal(
            sp1.quantities.extrema(("gas", f"velocity_{ax}")),
            sp2.quantities.extrema(("gas", f"velocity_{ax}")),
        )
    assert_allclose_units(
        sp1.quantities.bulk_velocity(), sp2.quantities.bulk_velocity()
    )
    assert_equal(prj1[("gas", "density")], prj2[("gas", "density")])


@requires_file(cloud)
def test_AthenaDataset():
    assert isinstance(data_dir_load(cloud), AthenaDataset)


@requires_file(sloshing)
@disable_dataset_cache
def test_mag_factor():
    ds1 = load(sloshing, units_override=uo_sloshing, magnetic_normalization="gaussian")
    assert ds1.magnetic_unit == np.sqrt(
        4.0 * np.pi * ds1.mass_unit / (ds1.time_unit**2 * ds1.length_unit)
    )
    sp1 = ds1.sphere("c", (100.0, "kpc"))
    pB1a = (
        sp1["athena", "cell_centered_B_x"] ** 2
        + sp1["athena", "cell_centered_B_y"] ** 2
        + sp1["athena", "cell_centered_B_z"] ** 2
    ) / (8.0 * np.pi)
    pB1b = (
        sp1["gas", "magnetic_field_x"] ** 2
        + sp1["gas", "magnetic_field_y"] ** 2
        + sp1["gas", "magnetic_field_z"] ** 2
    ) / (8.0 * np.pi)
    pB1a.convert_to_units("dyn/cm**2")
    pB1b.convert_to_units("dyn/cm**2")
    assert_allclose_units(pB1a, pB1b)
    assert_allclose_units(pB1a, sp1["magnetic_pressure"])
    ds2 = load(
        sloshing, units_override=uo_sloshing, magnetic_normalization="lorentz_heaviside"
    )
    assert ds2.magnetic_unit == np.sqrt(
        ds2.mass_unit / (ds2.time_unit**2 * ds2.length_unit)
    )
    sp2 = ds2.sphere("c", (100.0, "kpc"))
    pB2a = (
        sp2["athena", "cell_centered_B_x"] ** 2
        + sp2["athena", "cell_centered_B_y"] ** 2
        + sp2["athena", "cell_centered_B_z"] ** 2
    ) / 2.0
    pB2b = (
        sp2["gas", "magnetic_field_x"] ** 2
        + sp2["gas", "magnetic_field_y"] ** 2
        + sp2["gas", "magnetic_field_z"] ** 2
    ) / 2.0
    pB2a.convert_to_units("dyn/cm**2")
    pB2b.convert_to_units("dyn/cm**2")
    assert_allclose_units(pB2a, pB2b)
    assert_allclose_units(pB2a, sp2["magnetic_pressure"])
    assert_allclose_units(pB1a, pB2a)



"""
Shareable definitions for common fp/int Cython utilities



"""


cimport cython
cimport numpy as np


cdef inline np.int64_t imax(np.int64_t i0, np.int64_t i1) noexcept nogil:
    if i0 > i1: return i0
    return i1

cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) noexcept nogil:
    if f0 > f1: return f0
    return f1

cdef inline np.int64_t imin(np.int64_t i0, np.int64_t i1) noexcept nogil:
    if i0 < i1: return i0
    return i1

cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) noexcept nogil:
    if f0 < f1: return f0
    return f1

cdef inline np.float64_t fabs(np.float64_t f0) noexcept nogil:
    if f0 < 0.0: return -f0
    return f0

cdef inline np.int64_t iclip(np.int64_t i, np.int64_t a, np.int64_t b) noexcept nogil:
    if i < a: return a
    if i > b: return b
    return i

cdef inline np.int64_t i64clip(np.int64_t i, np.int64_t a, np.int64_t b) noexcept nogil:
    if i < a: return a
    if i > b: return b
    return i

cdef inline np.float64_t fclip(np.float64_t f,
                      np.float64_t a, np.float64_t b) noexcept nogil:
    return fmin(fmax(f, a), b)

cdef inline np.int64_t i64max(np.int64_t i0, np.int64_t i1) noexcept nogil:
    if i0 > i1: return i0
    return i1

cdef inline np.int64_t i64min(np.int64_t i0, np.int64_t i1) noexcept nogil:
    if i0 < i1: return i0
    return i1

cdef inline _ensure_code(arr):
    if hasattr(arr, "units"):
        if "code_length" == str(arr.units):
            return arr
        arr.convert_to_units("code_length")
    return arr

ctypedef fused any_float:
    np.float32_t
    np.float64_t



import numpy as np

import yt

ds = yt.load("MOOSE_sample_data/out.e-s010")
sc = yt.create_scene(ds)
cam = sc.camera

# save an image at the starting position
frame = 0
sc.save("camera_movement_%04i.png" % frame)
frame += 1

# Zoom out by a factor of 2 over 5 frames
for _ in cam.iter_zoom(0.5, 5):
    sc.save("camera_movement_%04i.png" % frame)
    frame += 1

# Move to the position [-10.0, 10.0, -10.0] over 5 frames
pos = ds.arr([-10.0, 10.0, -10.0], "code_length")
for _ in cam.iter_move(pos, 5):
    sc.save("camera_movement_%04i.png" % frame)
    frame += 1

# Rotate by 180 degrees over 5 frames
for _ in cam.iter_rotate(np.pi, 5):
    sc.save("camera_movement_%04i.png" % frame)
    frame += 1



from inspect import signature
from typing import Callable

import numpy as np

from yt.utilities.lib.misc_utilities import obtain_position_vector


def get_radius(data, field_prefix, ftype):
    center = data.get_field_parameter("center").to("code_length")
    DW = (data.ds.domain_right_edge - data.ds.domain_left_edge).to("code_length")
    # This is in code_length so it can be the destination for our r later.
    radius2 = data.ds.arr(
        np.zeros(data[ftype, field_prefix + "x"].shape, dtype="float64"), "code_length"
    )

    r = np.empty_like(radius2, subok=False)
    if any(data.ds.periodicity):
        rdw = radius2.v
    for i, ax in enumerate("xyz"):
        pos = data[ftype, f"{field_prefix}{ax}"]
        if str(pos.units) != "code_length":
            pos = pos.to("code_length")
        np.subtract(
            pos.d,
            center[i].d,
            r,
        )
        if data.ds.periodicity[i]:
            np.abs(r, r)
            np.subtract(r, DW.d[i], rdw)
            np.abs(rdw, rdw)
            np.minimum(r, rdw, r)
        np.multiply(r, r, r)
        np.add(radius2.d, r, radius2.d)
        if data.ds.dimensionality < i + 1:
            break
    # Using the views into the array is not changing units and as such keeps
    # from having to do symbolic manipulations
    np.sqrt(radius2.d, radius2.d)
    # Alias it, just for clarity.
    radius = radius2
    return radius


def get_periodic_rvec(data):
    coords = obtain_position_vector(data).d
    if sum(data.ds.periodicity) == 0:
        return coords
    le = data.ds.domain_left_edge.in_units("code_length").d
    dw = data.ds.domain_width.in_units("code_length").d
    for i in range(coords.shape[0]):
        if not data.ds.periodicity[i]:
            continue
        coords[i, ...] -= le[i]
        # figure out which measure is less
        mins = np.argmin(
            [
                np.abs(np.mod(coords[i, ...], dw[i])),
                np.abs(np.mod(coords[i, ...], -dw[i])),
            ],
            axis=0,
        )
        temp_coords = np.mod(coords[i, ...], dw[i])

        # Where second measure is better, updating temporary coords
        ii = mins == 1
        temp_coords[ii] = np.mod(coords[i, ...], -dw[i])[ii]

        # Putting the temporary coords into the actual storage
        coords[i, ...] = temp_coords

        coords[i, ...] + le[i]

    return coords


def validate_field_function(function: Callable) -> None:
    """
    Inspect signature, raise a TypeError if invalid, return None otherwise.
    """
    # This is a helper function to user-intended field registration methods
    # (e.g. Dataset.add_field and yt.derived_field)
    # it is not used in FieldInfoContainer.add_field to optimize performance
    # (inspect.signature is quite expensive and we don't want to validate yt's
    # internal code every time a dataset's fields are defined).

    # lookup parameters that do not have default values
    fparams = signature(function).parameters
    nodefaults = tuple(p.name for p in fparams.values() if p.default is p.empty)
    if nodefaults != ("field", "data"):
        raise TypeError(
            f"Received field function {function} with invalid signature. "
            f"Expected exactly 2 positional parameters ('field', 'data'), got {nodefaults!r}"
        )
    if any(
        fparams[name].kind == fparams[name].KEYWORD_ONLY for name in ("field", "data")
    ):
        raise TypeError(
            f"Received field function {function} with invalid signature. "
            "Parameters 'field' and 'data' must accept positional values "
            "(they cannot be keyword-only)"
        )
